"""Example how to generate, modify and plot impedance objects."""
import pysignalscope as pss

# import data as a channel object
example_data_0mm5 = pss.Impedance.from_waynekerr('impedance_example_data_0mm5.csv', '0.5mm air gap')
example_data_1mm5 = pss.Impedance.from_waynekerr('impedance_example_data1mm5.csv', '1.5mm air gap')
example_data_rlc = pss.Impedance.from_rlc('l', 1000, 500e-6, 10e-12)

# modify the data, e.g. change the color.
# For color changes, you can use standard colors or colors from the leapythontoolbox, see this example
example_data_0mm5 = pss.Impedance.modify(example_data_0mm5, color="red", unit="random unit")
example_data_1mm5 = pss.Impedance.modify(example_data_1mm5, color=pss.gnome_colors["blue"])
example_data_rlc = pss.Impedance.modify(example_data_rlc, color=pss.gnome_colors["green"], label="from rlc")

# # recalculate rlc data from a impedance curve
recalculated_r, recalculated_l, recalculated_c = (pss.Impedance.calc_rlc(example_data_rlc, 'l', f_calc_l=10e3, f_calc_c=10e7, plot_figure=True))
print(f"{recalculated_r=}")
print(f"{recalculated_l=}")
print(f"{recalculated_c=}")

# plot multiple channel data
impedance_figure = pss.Impedance.plot_impedance([example_data_0mm5, example_data_1mm5, example_data_rlc])
pss.save_figure(impedance_figure, "impedance_fig.pdf")

# save and load an impedance object
pss.Impedance.save(example_data_rlc, "example")
loaded_impedance_object = pss.Impedance.load("example.pkl")



"""Generate, modify and plot Impedance objects."""
# python libraries
from typing import Union, List, Tuple, Optional
import copy
import cmath
import os
import pickle

# 3rd party libraries
import numpy as np
from numpy import typing as npt
import matplotlib.pyplot as plt

# own libraries
from pysignalscope.impedance_dataclass import ImpedanceChannel

supported_measurement_devices = ['waynekerr', 'agilent']

class Impedance:
    """Class to share scope figures in a special format, to keep labels, units and voltages belonging to a certain curve."""

    unit_separator_plot = "/"

    @staticmethod
    def generate_impedance_object(frequency: Union[List, npt.ArrayLike], impedance: Union[List, npt.ArrayLike],
                                  phase: Union[List, npt.ArrayLike],
                                  label: Optional[str] = None, unit: Optional[str] = None, color: Union[str, tuple, None] = None,
                                  source: Optional[str] = None, linestyle: Optional[str] = None) -> ImpedanceChannel:
        """
        Generate the ImpedanceChannel object.

        :param frequency: channel frequency in Hz
        :type frequency: Union[List, npt.ArrayLike]
        :param impedance: channel impedance in Ohm
        :type impedance: Union[List, npt.ArrayLike]
        :param phase: channel phase in degree
        :type phase: Union[List, npt.ArrayLike]
        :param label: channel label to show in plots
        :type label: str
        :param unit: channel unit to show in plots
        :type unit: str
        :param color: channel color
        :type color: str
        :param source: Source, e.g. Measurement xy, Device yy
        :type source: str
        :param linestyle: line style for the plot e.g. '--'
        :type linestyle: str
        :return: ImpedanceChannel
        :rtype: ImpedanceChannel
        """
        if isinstance(frequency, List):
            frequency = np.array(frequency)
        elif isinstance(frequency, np.ndarray):
            frequency = frequency
        else:
            raise TypeError("frequency must be type list or ArrayLike")
        if isinstance(impedance, List):
            impedance = np.array(impedance)
        elif isinstance(impedance, np.ndarray):
            impedance = impedance
        else:
            raise TypeError("impedance must be type list or ArrayLike")
        if isinstance(phase, List):
            phase = np.array(phase)
        elif isinstance(phase, np.ndarray):
            phase = phase
        else:
            raise TypeError("phase must be type list or ArrayLike")

        # check for single non-allowed values in impedance
        if np.any(np.isnan(impedance)):
            raise ValueError("NaN is not allowed in impedance.")
        if np.any(np.isinf(impedance)):
            raise ValueError("inf is not allowed in impedance.")

        # check for single non-allowed values in phase
        if np.any(np.isnan(phase)):
            raise ValueError("NaN is not allowed in phase.")
        if np.any(np.isinf(phase)):
            raise ValueError("inf is not allowed in phase.")
        # check for empty data
        if frequency.size == 0:
            raise ValueError("Not allowed: frequency is empty")
        if impedance.size == 0:
            raise ValueError("Not allowed: impedance is empty")
        if phase.size == 0:
            raise ValueError("Not allowed: phase is empty")

        # check if frequency and impedance have the same length
        if len(frequency) != len(impedance):
            raise ValueError("frequency and impedance must be same length.")
        # check if frequency and phase have the same length
        if len(frequency) != len(phase):
            raise ValueError("frequency and phase must be same length.")

        # check if channel_time is strictly increasing
        if not np.all(np.diff(frequency) > 0):
            raise ValueError("frequency not strictly increasing.")

        # check label for a valid type
        if isinstance(label, str) or label is None:
            label = label
        else:
            raise TypeError("label must be type str or None.")
        # check channel unit for a valid type
        if isinstance(unit, str) or unit is None:
            unit = unit
        else:
            raise TypeError("unit must be type str or None.")
        # check color for a valid type
        if isinstance(color, str) or color is None:
            color = color
        else:
            raise TypeError("color must be type str or None.")
        # check source for a valid type
        if isinstance(source, str) or source is None:
            source = source
        else:
            raise TypeError("source must be type str or None.")
        # check linestyle for a valid type
        if isinstance(linestyle, str) or linestyle is None:
            linestyle = linestyle
        else:
            raise TypeError("linestyle must be type str or None.")

        return ImpedanceChannel(
            frequency=frequency,
            impedance=impedance,
            phase_deg=phase,
            label=label,
            unit=unit,
            color=color,
            source=source,
            linestyle=linestyle)

    @staticmethod
    def modify(channel: ImpedanceChannel, impedance_factor: Optional[float] = None, impedance_offset: Optional[float] = None,
               label: Optional[str] = None, unit: Optional[str] = None, color: Union[str, Tuple, None] = None,
               source: Optional[str] = None, linestyle: Optional[str] = None,
               frequency_cut_min: Optional[float] = None, frequency_cut_max: Optional[float] = None) -> ImpedanceChannel:
        """
        Modify channel data like metadata or add a factor or offset to channel data.

        Useful for ImpedanceChannel with frequency/data, but without labels or units.

        :param channel: Impedance object to modify
        :type channel: ImpedanceChannel
        :param impedance_factor: multiply channel.impedance by impedance_factor
        :type impedance_factor: float
        :param impedance_offset: add an offset to channel.impedance
        :type impedance_offset: float
        :param label: label to add to the Channel-class
        :type label: str
        :param unit: unit to add to the Channel-class
        :type unit: str
        :param color: Color of a channel
        :type color: str
        :param source: Source of a channel, e.g. 'GeckoCIRCUITS', 'Numpy', 'Tektronix-Scope', ...
        :type source: str
        :param frequency_cut_min: minimum frequency
        :type frequency_cut_min: float
        :param frequency_cut_max: maximum frequency
        :type frequency_cut_max: float
        :param linestyle: linestyle of channel, e.g. '--'
        :type linestyle: str
        :return: Modified ImpedanceChannel object
        :rtype: ImpedanceChannel
        """
        # deep copy to not modify the original input data
        modified_channel = copy.deepcopy(channel)

        if label is not None:
            modified_channel.label = label
        if unit is not None:
            modified_channel.unit = unit
        if impedance_factor is not None:
            modified_channel.impedance = modified_channel.impedance * impedance_factor
        if impedance_offset is not None:
            modified_channel.impedance = modified_channel.impedance + impedance_offset
        if color is not None:
            modified_channel.color = color
        if source is not None:
            modified_channel.source = source
        if linestyle is not None:
            modified_channel.linestyle = linestyle

        if frequency_cut_min is not None:
            index_list_to_remove = []
            for count, value in enumerate(modified_channel.frequency):
                if value < frequency_cut_min:
                    index_list_to_remove.append(count)
            modified_channel.frequency = np.delete(modified_channel.frequency, index_list_to_remove)
            modified_channel.impedance = np.delete(modified_channel.impedance, index_list_to_remove)
            modified_channel.phase_deg = np.delete(modified_channel.phase_deg, index_list_to_remove)

        if frequency_cut_max is not None:
            index_list_to_remove = []
            for count, value in enumerate(modified_channel.frequency):
                if value > frequency_cut_max:
                    index_list_to_remove.append(count)
            modified_channel.frequency = np.delete(modified_channel.frequency, index_list_to_remove)
            modified_channel.impedance = np.delete(modified_channel.impedance, index_list_to_remove)
            modified_channel.phase_deg = np.delete(modified_channel.phase_deg, index_list_to_remove)

        return modified_channel

    @staticmethod
    def copy(channel: ImpedanceChannel) -> ImpedanceChannel:
        """
        Create a deepcopy of ImpedanceChannel.

        :param channel: Impedance object
        :type channel: ImpedanceChannel
        :return: Deepcopy of the impedance object
        :rtype: ImpedanceChannel
        """
        if not isinstance(channel, ImpedanceChannel):
            raise TypeError("channel must be type ImpedanceChannel.")
        return copy.deepcopy(channel)

    @staticmethod
    def from_waynekerr(csv_filename: str, label: Optional[str] = None) -> 'ImpedanceChannel':
        """
        Bring csv-data from wayne kerr 6515b to Impedance.

        :param csv_filename: .csv filename from impedance analyzer
        :type csv_filename: str
        :param label: label to add to the Channel-class, optional.
        :type label: str
        :return: ImpedanceChannel object
        :rtype: ImpedanceChannel
        """
        impedance_measurement = np.genfromtxt(csv_filename, delimiter=',', dtype=float, skip_header=1,
                                              encoding='latin1')
        # move data to variables
        frequency = impedance_measurement[:, 0]
        impedance = impedance_measurement[:, 1]
        phase = impedance_measurement[:, 2]

        return Impedance.generate_impedance_object(frequency=frequency, impedance=impedance, phase=phase,
                                                   source='Impedance Analyzer Wayne Kerr 6515b', label=label)

    @staticmethod
    def from_kemet_ksim(csv_filename: str) -> 'ImpedanceChannel':
        """
        Import data from kemet "ksim" tool.

        :param csv_filename: path to csv-file
        :type csv_filename: str
        :return: ImpedanceChannel object
        :rtype: ImpedanceChannel
        """
        impedance_measurement = np.genfromtxt(csv_filename, delimiter=',', dtype=float, skip_header=1,
                                              encoding='latin1')
        # move data to variables
        frequency = impedance_measurement[:, 0]
        impedance = impedance_measurement[:, 3]
        phase = impedance_measurement[:, 4]

        return Impedance.generate_impedance_object(frequency=frequency, impedance=impedance, phase=phase,
                                                   source='https://ksim3.kemet.com/capacitor-simulation')

    @staticmethod
    def plot_impedance(channel_list: List, figure_size: Tuple = None) -> plt.figure:
        """
        Plot and compare impedance channels.

        :param channel_list: List with impedances
        :type channel_list: List
        :param figure_size: figure size as tuple in mm, e.g. (80, 80)
        :type figure_size: Tuple
        :return: matplotlib figure
        :rtype: plt.figure
        """
        fig, (ax1, ax2) = plt.subplots(nrows=2, ncols=1, sharex=True, figsize=[x / 25.4 for x in figure_size] if figure_size is not None else None)
        for channel in channel_list:
            ax1.loglog(channel.frequency, channel.impedance, label=channel.label,
                       color=channel.color, linestyle=channel.linestyle)
            ax2.semilogx(channel.frequency, channel.phase_deg, label=channel.label,
                         color=channel.color, linestyle=channel.linestyle)

        ax1.grid()
        ax1.legend()
        ax1.set(xlabel=f'Frequency {Impedance.unit_separator_plot} Hz', ylabel=rf'Impedance {Impedance.unit_separator_plot} $\Omega$')

        ax2.grid()
        ax2.set(xlabel=f'Frequency {Impedance.unit_separator_plot} Hz', ylabel=f'Phase {Impedance.unit_separator_plot} °')
        ax2.legend()
        plt.tight_layout()
        plt.show()
        return fig

    @staticmethod
    def plot_inductance_and_ac_resistance(channel_list: List) -> None:
        """
        Plot and compare inductance (in uH) and ac_resistance (Ohm) of impedance channels.

        :param channel_list: List with impedances
        :type channel_list: List
        """
        fig, (ax1, ax2) = plt.subplots(nrows=2, ncols=1, sharex=True)
        for channel in channel_list:
            ax1.semilogx(channel.frequency,
                         1e6 * channel.impedance * np.sin(np.deg2rad(channel.phase_deg)) / channel.frequency / 2 / np.pi,
                         label=channel.label, color=channel.color, linestyle=channel.linestyle)
            ax2.semilogx(channel.frequency,
                         channel.impedance * np.cos(np.deg2rad(channel.phase_deg)),
                         label=channel.label, color=channel.color, linestyle=channel.linestyle)

        ax1.grid()
        ax1.legend()
        ax1.set(xlabel=f'Frequency {Impedance.unit_separator_plot} Hz', ylabel=f'Inductance {Impedance.unit_separator_plot} ' + r'$\rm{\mu}$H')

        ax2.grid()
        ax2.set(xlabel=f'Frequency {Impedance.unit_separator_plot} Hz', ylabel=rf'AC Resistance {Impedance.unit_separator_plot} $\Omega$')
        ax2.legend()
        plt.show()

    @staticmethod
    def calc_re_im_parts(channel: ImpedanceChannel, show_figure: bool = True):
        """
        Calculate real and imaginary part of Impedance measurement.

        :param channel: Impedance object
        :type channel: ImpedanceChannel
        :param show_figure: Plot figure if true
        :type show_figure: bool
        :return: List with [(frequency, frequency_real_part), (frequency, frequency_imag_part)]
        :rtype: List
        """
        frequency_real_part = []
        frequency_imag_part = []
        complex_impedance = []

        for count_frequency, _ in enumerate(channel.frequency):
            impedance = channel.impedance[count_frequency] * np.exp(1j * channel.phase_deg[count_frequency] * 2 * np.pi / 360)
            complex_impedance.append(impedance)
            frequency_real_part.append(np.real(complex_impedance[count_frequency]))
            frequency_imag_part.append(np.imag(complex_impedance[count_frequency]))

        if show_figure:
            fig, (ax1, ax2) = plt.subplots(nrows=2, ncols=1, sharex=True)
            ax1.loglog(channel.frequency, frequency_real_part)
            ax1.grid()
            ax1.set_xlabel('frequency in Hz')
            ax1.set_ylabel('real (impedance)')

            ax2.loglog(channel.frequency, frequency_imag_part)
            ax2.set_xlabel('frequency in Hz')
            ax2.set_ylabel('imag (impedance)')
            ax2.grid()
            fig.show()

        frequency = copy.deepcopy(channel.frequency)

        return [(frequency, frequency_real_part), (frequency, frequency_imag_part)]

    @staticmethod
    def calc_rlc(channel: ImpedanceChannel, type_rlc: str, f_calc_c: float, f_calc_l: float, plot_figure: bool = False) -> tuple:
        """
        Calculate R, L, C values for given ImpedanceChannel.

        Calculated values will be drawn in a plot for comparison with the given data.

        :param channel: Impedance channel object
        :type channel: ImpedanceChannel
        :param type_rlc: Type 'R', 'L', 'C'
        :type type_rlc: str
        :param f_calc_c: Choose the frequency for calculation of C-value
        :type f_calc_c: float
        :param f_calc_l: Choose the frequency for calculation of L-value
        :type f_calc_l: float
        :param plot_figure: True/False [default] to plot the figure
        :type plot_figure: bool

        :return: Values for R, L, C
        :rtype: tuple

        :Example:

        >>> import pysignalscope as pss
        >>> example_data_rlc = pss.Impedance.from_rlc('l', 1000, 500e-6, 10e-12)
        >>> recalculated_r, recalculated_l, recalculated_c = pss.Impedance.calc_rlc(example_data_rlc, 'l', f_calc_l=10e3, f_calc_c=10e7, plot_figure=True)
        """
        # # Calculate R, L, C
        z_calc_c = np.interp(f_calc_c, channel.frequency, channel.impedance)
        z_calc_l = np.interp(f_calc_l, channel.frequency, channel.impedance)
        phase_calc_c = np.interp(f_calc_c, channel.frequency, channel.phase_deg)
        phase_calc_l = np.interp(f_calc_l, channel.frequency, channel.phase_deg)

        c_calc = 1 / (2 * np.pi * f_calc_c * z_calc_c)
        l_calc = z_calc_l / (2 * np.pi * f_calc_l)

        # # Calculate R at resonance frequency
        f_calc_r = 1 / (2 * np.pi * np.sqrt(l_calc * c_calc))
        z_calc_r = np.interp(f_calc_r, channel.frequency, channel.impedance)
        phase_calc_r = np.interp(f_calc_r, channel.frequency, channel.phase_deg)
        r_calc = z_calc_r

        if plot_figure:
            # Display calculated values
            print('C = {} F'.format(np.format_float_scientific(c_calc, precision=3)))
            print('R = {} Ohm'.format(np.format_float_scientific(r_calc, precision=3)))
            print('L = {} H'.format(np.format_float_scientific(l_calc, precision=3)))
            print('f_res = {} Hz'.format(np.format_float_scientific(f_calc_r, precision=1)))

            markerstyle = 'x'
            color_measurement = 'r'
            color_calculation = 'b'
            linestyle_measurement = '-'
            linestyle_calculation = ':'

            # plot output figure, compare measurement and interpolated data
            recalculated_curve = Impedance.from_rlc(type_rlc, r_calc, l_calc, c_calc)

            # generate plot
            fig, (ax1, ax2) = plt.subplots(nrows=2, ncols=1, sharex=True)

            # subplot 1 for impedance
            ax1.loglog(channel.frequency, channel.impedance, linestyle=linestyle_measurement, color=color_measurement, label='measurement')
            ax1.loglog(recalculated_curve.frequency, abs(recalculated_curve.impedance), linestyle=linestyle_calculation,
                       color=color_calculation, label='recalculated data')
            ax1.grid()
            ax1.legend()
            ax1.set(xlabel=f'Frequency {Impedance.unit_separator_plot} Hz', ylabel=rf'Impedance {Impedance.unit_separator_plot} $\Omega$')
            ax1.plot(f_calc_c, z_calc_c, marker=markerstyle, color=color_measurement)
            ax1.plot(f_calc_r, z_calc_r, marker=markerstyle, color=color_measurement)
            ax1.plot(f_calc_l, z_calc_l, marker=markerstyle, color=color_measurement)

            # subplot 2 for phase
            ax2.semilogx(channel.frequency, channel.phase_deg, linestyle=linestyle_measurement, color=color_measurement, label='measurement')
            ax2.semilogx(recalculated_curve.frequency, recalculated_curve.phase_deg, linestyle=linestyle_calculation, color=color_calculation,
                         label='recalculated data')
            ax2.grid()
            ax2.set(xlabel=f'Frequency {Impedance.unit_separator_plot} Hz', ylabel=f'Phase {Impedance.unit_separator_plot} °')
            ax2.legend()
            ax2.plot(f_calc_c, phase_calc_c, marker=markerstyle, color=color_measurement)
            ax2.plot(f_calc_r, phase_calc_r, marker=markerstyle, color=color_measurement)
            ax2.plot(f_calc_l, phase_calc_l, marker=markerstyle, color=color_measurement)
            plt.show()

        # return results
        return r_calc, l_calc, c_calc

    @staticmethod
    def from_rlc(type_rlc: str, resistance: Union[float, np.array], inductance: Union[float, np.array],
                 capacitance: Union[float, np.array]) -> 'ImpedanceChannel':
        """
        Calculate the impedance over frequency for R - L - C - combination.

        :param type_rlc: Type of network, can be 'R', 'L' or 'C'
        :type type_rlc: str
        :param resistance: resistance
        :type resistance: float
        :param inductance: inductance
        :type inductance: float
        :param capacitance: capacitance
        :type capacitance: bool
        :return: Impedance object
        :rtype: ImpedanceChannel

        :Example:

        >>> import pysignalscope as pss
        >>> impedance_channel_object = pss.Impedance.from_rlc('C', 10e-3, 100e-9, 36e-3)

         *  Type C and RLC

         .. code-block::

            ---R---L---C---

         *  Type R

         .. code-block::

            ---+---R---L---+---
               |           |
               +-----C-----+

         *  Type L

         .. code-block::

            ---+---L---+---
               |       |
               +---C---+
               |       |
               +---R---+
        """
        f = np.logspace(0, 8, 10000)

        if 'C' in type_rlc.upper() or 'RLC' in type_rlc.upper():
            z_total = resistance + 1j * 2 * np.pi * f * inductance - 1j / (2 * np.pi * f * capacitance)
        elif 'R' in type_rlc.upper():
            z_total = 1. / (1. / (resistance + 1j * 2 * np.pi * f * inductance) + 1j * 2 * np.pi * f * capacitance)
        elif 'L' in type_rlc.upper():
            z_total = 1 / (1. / resistance + 1. / (2 * np.pi * f * inductance * 1j) + 1j * 2 * np.pi * f * capacitance)
        else:
            raise ValueError("check input type_rlc!")

        phase_vector = []
        for i in z_total:
            phase_vector.append(cmath.phase(i) * 360 / (2 * np.pi))

        return Impedance.generate_impedance_object(frequency=f, impedance=np.abs(z_total), phase=phase_vector)

    @staticmethod
    def check_capacitor_from_waynekerr(csv_filename: str, label: str,
                                       target_capacitance: float, plot_figure: bool = True) -> 'ImpedanceChannel':
        """
        Check a capacitor impedance .csv-curve against a target capacitance.

        Reads the .csv-curve from wayne kerr impedance analyzer, calculates r, l, and c from the measurement
        and shows the derivation from the target capacitance. Visualizes the curves in case of plot_figure is set
        to true.

        :param csv_filename: filepath to Wayne Kerr impedance analyzer .csv-file
        :type csv_filename: str
        :param label: channel label for the plot
        :type label: str
        :param target_capacitance: target capacitance in F
        :type target_capacitance: float
        :param plot_figure: Set to True for plot
        :type plot_figure: bool

        :return: measured capacitor as impedance curve
        :rtype: ImpedanceChannel
        """
        f_calc_c = 100
        f_calc_l = 15e6

        capacitor_impedance = Impedance.from_waynekerr(csv_filename, label)

        r_measure, l_measure, c_measure = Impedance.calc_rlc(capacitor_impedance, "c", f_calc_c, f_calc_l, plot_figure=plot_figure)
        derivation_capacitance = c_measure / target_capacitance

        print(f"{derivation_capacitance=}")

        return capacitor_impedance

    @staticmethod
    def save(channel: ImpedanceChannel, filepath: str) -> None:
        """
        Save an ImpedanceChannel object to hard disk.

        :param channel: impedance channel
        :type channel: ImpedanceChannel
        :param filepath: filepath including file name
        :type filepath: str
        """
        if not isinstance(filepath, str):
            raise TypeError("filepath must be of type str.")
        if ".pkl" not in filepath:
            filepath = filepath + ".pkl"
        file_path, file_name = os.path.split(filepath)
        if file_path == "":
            file_path = os.path.curdir
        if not os.path.exists(file_path):
            os.makedirs(file_path, exist_ok=True)
        if not isinstance(channel, ImpedanceChannel):
            raise TypeError("channel must be of type ImpedanceChannel.")

        with open(filepath, 'wb') as handle:
            pickle.dump(channel, handle, protocol=pickle.HIGHEST_PROTOCOL)

    @staticmethod
    def load(filepath: str) -> ImpedanceChannel:
        """
        Load an ImpedanceChannel file from the hard disk.

        :param filepath: filepath
        :type filepath: str
        :return: loaded ImpedanceChannel object
        :rtype: ImpedanceChannel
        """
        if not isinstance(filepath, str):
            raise TypeError("filepath must be of type str.")
        if ".pkl" not in filepath:
            raise ValueError("filepath must end with .pkl")
        if not os.path.exists(filepath):
            raise ValueError(f"{filepath} does not exist.")
        with open(filepath, 'rb') as handle:
            loaded_scope_object: ImpedanceChannel = pickle.load(handle)
        if not isinstance(loaded_scope_object, ImpedanceChannel):
            raise TypeError(f"Loaded object is of type {type(loaded_scope_object)}, but should be type Scope.")

        return loaded_scope_object

    @staticmethod
    def to_inductance(channel: ImpedanceChannel) -> ImpedanceChannel:
        """
        Convert an ImpedanceChannel to a pure inductance ImpedanceChannel.

        Ignore the resistive and capacitive part by trigonometric operation.

        :param channel: Impedance channel object
        :type channel: ImpedanceChannel
        :return: Impedance channel as inductance
        :rtype: ImpedanceChannel
        """
        if not isinstance(channel, ImpedanceChannel):
            raise TypeError("channel must be type ImpedanceChannel.")
        inductance_phase = 90 * np.ones_like(channel.phase_deg)
        inductance_impedance = channel.impedance * np.sin(np.deg2rad(channel.phase_deg))

        inductance_channel = Impedance.generate_impedance_object(channel.frequency, impedance=inductance_impedance, phase=inductance_phase)

        return inductance_channel

    @staticmethod
    def to_resistance(channel: ImpedanceChannel) -> ImpedanceChannel:
        """
        Convert an ImpedanceChannel to a pure resistive ImpedanceChannel.

        Ignore the inductive and capacitive part by trigonometric operation.

        :param channel: Impedance channel object
        :type channel: ImpedanceChannel
        :return: Impedance channel as inductance
        :rtype: ImpedanceChannel
        """
        if not isinstance(channel, ImpedanceChannel):
            raise TypeError("channel must be type ImpedanceChannel.")
        resistance_phase = np.zeros_like(channel.phase_deg)
        resistance_impedance = channel.impedance * np.cos(np.deg2rad(channel.phase_deg))

        resistance_channel = Impedance.generate_impedance_object(channel.frequency, impedance=resistance_impedance, phase=resistance_phase)

        return resistance_channel

    @staticmethod
    def to_capacitance(channel: ImpedanceChannel) -> ImpedanceChannel:
        """
        Convert an ImpedanceChannel to a pure capacitance ImpedanceChannel.

        Ignore the resistive and inductive part by trigonometric operation.

        :param channel: Impedance channel object
        :type channel: ImpedanceChannel
        :return: Impedance channel as inductance
        :rtype: ImpedanceChannel
        """
        if not isinstance(channel, ImpedanceChannel):
            raise TypeError("channel must be type ImpedanceChannel.")
        capacitance_phase = -90 * np.ones_like(channel.phase_deg)
        capacitance_impedance = -channel.impedance * np.sin(np.deg2rad(channel.phase_deg))

        capacitance_channel = Impedance.generate_impedance_object(channel.frequency, impedance=capacitance_impedance, phase=capacitance_phase)

        return capacitance_channel

    @staticmethod
    def plot_component(channel: ImpedanceChannel, figure_size: Optional[Tuple] = None) -> None:
        """
        Plot the component values.

        Phase must be -90°, 0°, +90° for all entries.

        :param channel: Impedance
        :type channel: List
        :param figure_size: figure size as tuple in mm, e.g. (80, 80)
        :type figure_size: Tuple
        """
        if not isinstance(channel, ImpedanceChannel):
            raise TypeError("channel must be type ImpedanceChannel.")

        fig, (ax1, ax2) = plt.subplots(nrows=2, ncols=1, sharex=True, figsize=[x / 25.4 for x in figure_size] if figure_size is not None else None)

        # type resistor
        if np.array_equal(channel.phase_deg, np.zeros_like(channel.phase_deg)):
            ax1.semilogx(channel.frequency, channel.impedance)
            ax1.set_ylabel(fr"Resistance {Impedance.unit_separator_plot} $\Omega$")

        # type capacitance
        elif np.array_equal(channel.phase_deg, -90 * np.ones_like(channel.phase_deg)):
            capacitance = 1 / channel.impedance / 2 / np.pi / channel.frequency
            capacitance[capacitance <= 0] = np.nan
            ax1.semilogx(channel.frequency, capacitance)
            ax1.set_ylabel(rf"Capacitance {Impedance.unit_separator_plot} F")

        # type inductance
        elif np.array_equal(channel.phase_deg, 90 * np.ones_like(channel.phase_deg)):
            inductance = channel.impedance / 2 / np.pi / channel.frequency
            inductance[inductance <= 0] = np.nan
            ax1.semilogx(channel.frequency, inductance)
            ax1.set_ylabel(fr"Inductance {Impedance.unit_separator_plot} H")
        else:
            raise ValueError("Can not detect type of component. Phase must be -90°, 0°, +90° for all entries.")

        ax1.set_xlabel(f"Frequency {Impedance.unit_separator_plot} Hz")
        ax1.grid()
        ax2.semilogx(channel.frequency, channel.phase_deg)
        ax2.grid()
        ax2.set_xlabel(f"Frequency {Impedance.unit_separator_plot} Hz")
        ax2.set_ylabel(f"Phase {Impedance.unit_separator_plot} °")

        plt.show()



"""Class supports shifting channels within displayed plot."""
from enum import Enum
from typing import List, Tuple
import os
# 3rd party libraries
from matplotlib import pyplot as plt
# own libraries
from pysignalscope.logconfig import setup_logging
from pysignalscope.scope_dataclass import Channel
# python libraries
import logging
# Interactive shift plot
import matplotlib
import matplotlib.patches as patches
from matplotlib.widgets import Button, TextBox
from matplotlib.widgets import Slider

# in case of GitHubs CI, use Agg instead of TkAgg (normal usage)
if "IS_TEST" in os.environ:
    matplotlib.use('Agg')
else:
    matplotlib.use('TkAgg')

# - Logging setup ---------------------------------------------------------------------------------
setup_logging()

# Modul name für static methods
class_modulename = "channelshift"

# - Class definition ------------------------------------------------------------------------------

class ScopeChShift:
    """Class to shift channels interactively."""

    # - private members ------------------------------------------------------------------------------

    # Index of selected channel
    chn_index = 0
    last_val = 0
    shift_dir = 1

    # Shiftfigure
    shiftfig = None
    # Zoom axe
    zoom_ax = None
    # Reference channels to plot
    channelplotlist = list()
    # List of shift of channels
    shiftlist = list()

    # Widget container variables
    shift_text_box = None
    shsl_reset_button = None
    chn_sel_button = None
    dir_shift_button = None
    shift_slider = None
    selplotlabel = None

    # Shift step variables
    shiftstep_x = None
    shiftstep_y = None
    # Limits of shift steps
    min_shiftstep_x = None
    max_shiftstep_x = None
    min_shiftstep_y = None
    max_shiftstep_y = None

    # Maximal values
    display_min_x = None
    display_max_x = None
    display_min_y = None
    display_max_y = None

    # Minimal zoom values
    zoom_delta_y = None
    zoom_delta_x = None

    # Shift zoom variables
    zoom_start_x = None
    zoom_start_y = None
    # Pixelbox for the graphic
    shiftfigbox = None
    # Zoomrect
    zoom_rect = None

    # Zoom state variabe

    class Zoom_State(Enum):
        """Enumeration to control the zoom state."""

        NoZoom = 0
        ZoomSelect = 1
        ZoomConfirm = 2

    zoom_state = Zoom_State.NoZoom

    @staticmethod
    def channel_shift(channels: List['Channel'], shiftstep_x: float, shiftstep_y: float,
                      displayrange_x: [float, float], displayrange_y: [float, float]):
        """
        Interactive plot of channel datasets.

        Examples:
        >>> import pysignalscope as pss
        >>> chshift : ScopeChShift()
        >>> ch1=pss.Scope.generate_channel([0, 5e-4, 10e-4, 15e-4, 20e-4],[20,150,130,140,2])
        >>> ch2=pss.Scope.generate_channel([10e-4, 12e-4, 21-4, 35e-4, 40e-4],[-20,130,-30,100,40])
        >>> chshift.init_shiftstep_limits((1e-6,5e-5),(1,20))
        >>> chshift.channel_shift([ch1,ch2], 1e-5,10,[0,40e-4],[-100,200])
        Plots the channels ch1 and ch2. You can zoom into by selecting the zoom area with help of
        left mouse button. By moving the mouse while pressing the button  the area is marked by a red rectangle.
        If you release the left mouse button the area is marked. By moving the mouse within the area an perform
        a button press, you confirm and you zoom in the area. If you perform the left mouse button click outside
        of the marked area, you reject the selection. You reject the selection always by clicking the right mouse button independent you zoom out.
        button. If no area is selected or wait for confirmation, the click on the right mouse button leads to zooming out.
        There is a zoom limit in both directions. In this case, the rectangle shows the possible area (becomes larger),
        after you have release the left mouse button.

        Y-axis labels are set according to the channel_unit, presented in the last curve for the subplot.
        For own axis labeling, use as channel_unit for the last channel your label, e.g. r"$i_T$ in A".
        Note, that the r before the string gives the command to accept LaTeX formulas, like $$.
        The parameters has to fullfill conditions:
        Minimal shift step in x-direction is the minimal difference of 2 points of all provided channels

        :param channels: list of datasets
        :type channels: list[Channel]
        :param shiftstep_x: shift step in x-direction
        :type shiftstep_x: float
        :param shiftstep_y: shift step in y-direction
        :type shiftstep_y: float
        :param displayrange_x: Display range limits in x-direction (min_x, max_x)
        :type displayrange_x: tuple of float
        :param displayrange_y: Display range limits in y-direction (min_y, max_y)
        :type displayrange_y: tuple of float

        :return: List of x and y-shifts per channel
        :rtype: list[float]
        """
        # Create plot and clear lists
        ScopeChShift.shiftfig, ScopeChShift.zoom_ax = plt.subplots()
        ScopeChShift.channelplotlist = list()
        ScopeChShift.shiftlist = list()
        # Actual display range (x and y)
        act_displayrange_x = [0.0, 0.0]
        act_displayrange_y = [0.0, 0.0]

        # Read channel data for the plot
        for channel in channels:
            cur_channelplot, = ScopeChShift.zoom_ax.plot(channel.time, channel.data, label=channel.label, color=channel.color)
            ScopeChShift.channelplotlist.append(cur_channelplot)
            ScopeChShift.shiftlist.append([0, 0])

        # Init minimum and maximum values of the display window if required
        ScopeChShift.display_min_x, ScopeChShift.display_max_x = ScopeChShift.zoom_ax.get_xlim()

        # Check, if displayrange x need to initialize
        if displayrange_x[0] == displayrange_x[1]:
            act_displayrange_x[0] = ScopeChShift.display_min_x
            act_displayrange_x[1] = ScopeChShift.display_max_x
        else:
            # Overtake values as display range
            act_displayrange_x[0] = displayrange_x[0]
            act_displayrange_x[1] = displayrange_x[1]
            # Overtake values as new display range limits, if the range becomes higher
            if act_displayrange_x[0] < ScopeChShift.display_min_x:
                ScopeChShift.display_min_x = act_displayrange_x[0]
            if act_displayrange_x[1] > ScopeChShift.display_max_x:
                ScopeChShift.display_max_x = act_displayrange_x[1]

        # Set display range in x-direction
        ScopeChShift.zoom_ax.set_xlim(act_displayrange_x[0], act_displayrange_x[1])

        # Init minimum and maximum values of the display window if required
        ScopeChShift.display_min_y, ScopeChShift.display_max_y = ScopeChShift.zoom_ax.get_ylim()

        # Check, if displayrange y need to initialize
        if displayrange_y[0] == displayrange_y[1]:
            act_displayrange_y[0] = ScopeChShift.display_min_y
            act_displayrange_y[1] = ScopeChShift.display_max_y
        else:
            # Overtake values as display range
            act_displayrange_y[0] = displayrange_y[0]
            act_displayrange_y[1] = displayrange_y[1]
            # Overtake values as new display range limits, if the range becomes higher
            if act_displayrange_y[0] < ScopeChShift.display_min_y:
                ScopeChShift.display_min_y = act_displayrange_y[0]
            if act_displayrange_y[1] > ScopeChShift.display_max_y:
                ScopeChShift.display_max_y = act_displayrange_y[1]

        # Set display range in y-direction
        ScopeChShift.zoom_ax.set_ylim(act_displayrange_y[0], act_displayrange_y[1])

        # Define minimum zoom window as min_shiftstep_[xy]*5
        ScopeChShift.zoom_delta_y = ScopeChShift.min_shiftstep_y * 5
        ScopeChShift.zoom_delta_x = ScopeChShift.min_shiftstep_x * 5

        # Reset  channel index
        ScopeChShift.chn_index = 0
        # Set shift direction 0=x, 1=y
        ScopeChShift.shift_dir = 1
        # Set slider value
        ScopeChShift.last_val = 0
        # Hint for the user
        ScopeChShift.zoom_ax.set_title("Move the slider to move the selected channel curve")
        plt.subplots_adjust(bottom=0.45)

        # Overtake shift step parameter
        ScopeChShift.shiftstep_x = shiftstep_x
        ScopeChShift.shiftstep_y = shiftstep_y

        # -- Create the widgets

        # Button for selection of shift direction (x or y)
        button_ax = plt.axes([0.1, 0.15, 0.3, 0.075])  # Position and size of button
        ScopeChShift.dir_shift_button = Button(button_ax, 'Shift y')
        ScopeChShift.dir_shift_button.on_clicked(ScopeChShift.__toggle_xy_plot)  # Link to callback

        # Button for selection of graph
        button_ax = plt.axes([0.1, 0.25, 0.3, 0.075])  # Position and size of button
        ScopeChShift.chn_sel_button = Button(button_ax, 'Sel. Plot')
        ScopeChShift.chn_sel_button.on_clicked(ScopeChShift.__next_channel)  # links button to function

        # Button for reset slider position
        button_ax = plt.axes([0.85, 0.1, 0.14, 0.1])  # Position and size of button
        ScopeChShift.shsl_reset_button = Button(button_ax, 'Re-calibrate\nslider')
        ScopeChShift.shsl_reset_button.on_clicked(ScopeChShift.__reset_slider)  # links button to function

        # Textbox for step size
        text_box_ax = plt.axes([0.6, 0.15, 0.2, 0.05])
        ScopeChShift.shift_text_box = TextBox(text_box_ax, 'Shiftstep:', initial=str(ScopeChShift.shiftstep_y))
        ScopeChShift.shift_text_box.on_submit(ScopeChShift.__submit)

        # Selected dataset (Show label)
        labeltext = ScopeChShift.channelplotlist[0].get_label()
        # Check, if label text is not set
        if labeltext is None:
            labeltext = "●"
        else:
            labeltext = "● " + labeltext
        # Set labeltext in figure
        ScopeChShift.selplotlabel = ScopeChShift.shiftfig.text(0.6, 0.3, labeltext, ha='left', va='top', fontsize=12)
        ScopeChShift.selplotlabel.set_color(ScopeChShift.channelplotlist[ScopeChShift.chn_index].get_color())

        # Register eventhandler
        ScopeChShift.shiftfig.canvas.mpl_connect('button_press_event', ScopeChShift.__on_press)
        ScopeChShift.shiftfig.canvas.mpl_connect('motion_notify_event', ScopeChShift.__on_motion)
        ScopeChShift.shiftfig.canvas.mpl_connect('button_release_event', ScopeChShift.__on_release)

        # Slider for shift method
        slider_ax = plt.axes([0.2, 0.1, 0.55, 0.03], facecolor="lightgoldenrodyellow")  # Position and size of sliders
        ScopeChShift.shift_slider = Slider(slider_ax, 'Shift', -10.0, 10.0, valinit=0)  # Slider values
        # Link slider with callback
        ScopeChShift.shift_slider.on_changed(ScopeChShift.__shiftchannel)
        # Get the figure box object for check purpose
        ScopeChShift.shiftfigbox = ScopeChShift.zoom_ax.get_window_extent()

        # Log flow control
        logging.debug(f"{class_modulename} :\nDisplay range x:{ScopeChShift.display_min_x},{ScopeChShift.display_max_x} "
                      f"Display range y:{ScopeChShift.display_min_y},{ScopeChShift.display_max_y}\n"
                      f"Shift step x min,act,max:{ScopeChShift.min_shiftstep_x}, {shiftstep_x},{ScopeChShift.max_shiftstep_x}\n"
                      f"Shift step y min,act,max:{ScopeChShift.min_shiftstep_y}, {shiftstep_y},{ScopeChShift.max_shiftstep_y}\n"
                      f"Zoom area x,y:{ScopeChShift.zoom_delta_x}, {ScopeChShift.zoom_delta_y}")

        # Shows the figure
        plt.show()

        # Return the list of channel shifts
        return ScopeChShift.shiftlist

    @staticmethod
    def init_shiftstep_limits(shiftsteprange_x: Tuple[float, float], shiftsteprange_y: Tuple[float, float]):
        """
        Initialize the minimal and maximal shift step.

        Examples:
        >>> import pysignalscope as pss
        >>> chshift : ScopeChShift()
        >>> chshift.init_shiftstep_limits((1e-6,5e-5),(1,20))
        Set the limit of the shift step

        :param shiftsteprange_x: Default shift step, in calsLimits of shift steps in x-direction (min_x, max_x)
        :type shiftsteprange_x: tuple of float    :param shiftsteprange_x: Limits of shift steps in x-direction (min_x, max_x)
        :type shiftsteprange_x: tuple of float
        :param shiftsteprange_y: Limits of shift steps in y-direction (min_y, max_y)
        :type shiftsteprange_y: tuple of float
        """
        # Overtake shiftsteprange in x-direction
        ScopeChShift.min_shiftstep_x = shiftsteprange_x[0]
        ScopeChShift.max_shiftstep_x = shiftsteprange_x[1]

        # Overtake shiftsteprange in y-direction
        ScopeChShift.min_shiftstep_y = shiftsteprange_y[0]
        ScopeChShift.max_shiftstep_y = shiftsteprange_y[1]

    ##############################################################################
    # Callback methods for interactive shifting of plots
    ##############################################################################

    @staticmethod
    def __next_channel(event: any):
        """
        Select the next channel.

        This callback method is private (only for internal usage)
        Called by matplotlib-Event and assigned to a button.

        :param event: Container with some information: x and y position in pixel, x and y position in data scale,...
        :type event: any
        """
        # Increment the index of the selected channel
        ScopeChShift.chn_index = ScopeChShift.chn_index + 1
        # Check on overflow
        if ScopeChShift.chn_index >= len(ScopeChShift.channelplotlist):
            ScopeChShift.chn_index = 0
        # Display the label
        labeltext = ScopeChShift.channelplotlist[ScopeChShift.chn_index].get_label()
        # Check, if label text is not set
        if labeltext is None:
            labeltext = "●"
        else:
            labeltext = "● " + labeltext

        ScopeChShift.selplotlabel.set_text(labeltext)
        ScopeChShift.selplotlabel.set_color(ScopeChShift.channelplotlist[ScopeChShift.chn_index].get_color())

        ScopeChShift.shiftfig.canvas.draw_idle()  # update view
        ScopeChShift.last_val = 0
        ScopeChShift.shift_slider.set_val(0)

        # Log flow control
        logging.debug(f"{class_modulename} :Channel number {ScopeChShift.chn_index} is selected.")

    @staticmethod
    def __toggle_xy_plot(event):
        """
        Toggle the shift direction.

        This callback method is private (only for internal usage)
        Called by matplotlib-Event and assigned to a button.

        :param event: Container with some information: x and y position in pixel, x and y position in data scale,...
        :type event: any
        """
        # Check shift direction
        if ScopeChShift.shift_dir == 1:
            # Toggle to x-direction
            ScopeChShift.shift_dir = 0
            # Set button text
            ScopeChShift.dir_shift_button.label.set_text("Shift x")
            ScopeChShift.shift_text_box.set_val(str(ScopeChShift.shiftstep_x))
        else:
            # Toggle to y-direction
            ScopeChShift.shift_dir = 1
            # Set button text
            ScopeChShift.dir_shift_button.label.set_text("Shift y")
            ScopeChShift.shift_text_box.set_val(str(ScopeChShift.shiftstep_y))
        # Reset the values
        ScopeChShift.last_val = 0
        ScopeChShift.shift_slider.set_val(0)
        # Update the plot
        ScopeChShift.shiftfig.canvas.draw_idle()

        # Log flow control
        logging.debug(f"{class_modulename} :Shift direction is toggled to {ScopeChShift.shift_dir} (0=x, 1=y).")

    @staticmethod
    def __submit(text: str):
        """
        Change the shift-step size.

        This callback method is private (only for internal usage)
        Called by matplotlib-Event and assigned to an update of text.

        :param text: Updated text
        :type text: string
        """
        # Check if text is a number
        try:
            fvalue = float(text)
        except ValueError:
            fvalue = 0
        # Check shift direction
        if ScopeChShift.shift_dir == 1:
            ScopeChShift.shiftstep_y = fvalue
            if ScopeChShift.shiftstep_y > ScopeChShift.max_shiftstep_y:
                # Shift step x too high
                logging.info(
                    f"{class_modulename} :Shift step y {ScopeChShift.shiftstep_y}"
                    f" too high and set to {ScopeChShift.max_shiftstep_y}")
                # Correct the value
                ScopeChShift.shiftstep_y = ScopeChShift.max_shiftstep_y
                # Log flow control
                logging.info(f"{class_modulename} :Shift step y too small.{ScopeChShift.shiftstep_y}")
            elif ScopeChShift.shiftstep_y < ScopeChShift.min_shiftstep_y:
                # Shift step x too small
                logging.info(
                    f"{class_modulename} :Shift step y {ScopeChShift.shiftstep_y}"
                    f" too small and set to {ScopeChShift.min_shiftstep_y}")
                # Correct the value
                ScopeChShift.shiftstep_y = ScopeChShift.min_shiftstep_y

            ScopeChShift.shift_text_box.set_val(f"{ScopeChShift.shiftstep_y:.7g}")
            # Log flow control
            logging.debug(f"{class_modulename} :Shift step y = {ScopeChShift.shiftstep_y}")

        else:
            ScopeChShift.shiftstep_x = fvalue
            if ScopeChShift.shiftstep_x > ScopeChShift.max_shiftstep_x:
                # Shift step x too high
                logging.info(
                    f"{class_modulename} :Shift step x {ScopeChShift.shiftstep_y} too high and set to {ScopeChShift.max_shiftstep_x}")
                # Correct the value
                ScopeChShift.shiftstep_x = ScopeChShift.max_shiftstep_x
            elif ScopeChShift.shiftstep_x < ScopeChShift.min_shiftstep_x:
                # Shift step x too small
                logging.info(
                    f"{class_modulename} :Shift step x {ScopeChShift.shiftstep_x} too small and set to {ScopeChShift.min_shiftstep_x}")
                # Correct the value
                ScopeChShift.shiftstep_x = ScopeChShift.min_shiftstep_x

            # Update the format if necessary
            ScopeChShift.shift_text_box.set_val(f"{ScopeChShift.shiftstep_x:.7g}")

        # Log flow control
        logging.debug(
            f"{class_modulename}: Shift step x = {ScopeChShift.shiftstep_x},Shift step y = {ScopeChShift.shiftstep_y}")

        # Reset the values
        ScopeChShift.last_val = 0
        ScopeChShift.shift_slider.set_val(0)
        # Update the plot
        ScopeChShift.shiftfig.canvas.draw_idle()

        # Log flow control
        logging.debug(f"{class_modulename} :Text is overtaken {ScopeChShift.shift_dir} (0=x, 1=y).")

    @staticmethod
    def __shiftchannel(val: float):
        """
        Shift the channel by movement of the slider.

        This callback method is private (only for internal usage)
        Called by matplotlib-Event and assigned to the slider

        :param val:  slider position
        :type val: float
        """
        # Check shift direction
        if ScopeChShift.shift_dir == 1:
            delta = val - ScopeChShift.last_val
            dshift = (delta * ScopeChShift.shiftstep_y)
            ScopeChShift.shiftlist[ScopeChShift.chn_index][ScopeChShift.shift_dir] = (
                ScopeChShift.shiftlist[ScopeChShift.chn_index][ScopeChShift.shift_dir] + dshift
            )
            # Shift values in y-direction
            new_y = [value + dshift for value in ScopeChShift.channelplotlist[ScopeChShift.chn_index].get_ydata()]
            # Update dataset
            ScopeChShift.channelplotlist[ScopeChShift.chn_index].set_ydata(new_y)
        else:
            delta = val - ScopeChShift.last_val
            dshift = (delta * ScopeChShift.shiftstep_x)
            shift = ScopeChShift.shiftlist[ScopeChShift.chn_index][ScopeChShift.shift_dir] + dshift
            # Shift in x-direction
            new_x = [value + dshift for value in ScopeChShift.channelplotlist[ScopeChShift.chn_index].get_xdata()]
            # Update dataset
            ScopeChShift.channelplotlist[ScopeChShift.chn_index].set_xdata(new_x)

            # Overtake the shift
            ScopeChShift.shiftlist[ScopeChShift.chn_index][ScopeChShift.shift_dir] = shift
            # Update the plot
            ScopeChShift.shiftfig.canvas.draw_idle()

        # Store  slider value
        ScopeChShift.last_val = val

    @staticmethod
    def __reset_slider(event: any):
        """
        Recalibrate the slider position to zero without movement of the channel.

        This method is private (only for internal usage)
        Called by matplotlib-Event and assigned to a button

        :param event: Container with some information: x and y position in pixel, x and y position in data scale,...
        :type event: any
        """
        ScopeChShift.last_val = 0
        ScopeChShift.shift_slider.set_val(0)
        # Update the plot
        ScopeChShift.shiftfig.canvas.draw_idle()

        # Log flow control
        logging.debug(f"{class_modulename} :Slider reset was called.")

    # -- Zoomcallbacks ------------------------------------------------------------------------------------

    @staticmethod
    def __on_press(event: any):
        """
        Notification event if a mouse button is pressed.

        This callback method is private (only for internal usage)
        Called by matplotlib-Event on button press

        :param event: Container with some information: x and y position in pixel, x and y position in data scale,...
        :type event: any
        """
        # Check, if the button press was within the area
        if event.inaxes and ScopeChShift.shiftfigbox.contains(event.x, event.y):
            # Check, if the left button was pressed
            if event.button == 1:
                if ScopeChShift.zoom_state == ScopeChShift.Zoom_State.NoZoom:
                    # Overtake values
                    ScopeChShift.zoom_start_x, ScopeChShift.zoom_start_y = event.xdata, event.ydata
                    # Create rectangle
                    ScopeChShift.zoom_rect = patches.Rectangle((ScopeChShift.zoom_start_x, ScopeChShift.zoom_start_y),
                                                               0, 0, linewidth=1, edgecolor='red', facecolor='none')
                    # Set Zoomstate
                    ScopeChShift.zoom_state = ScopeChShift.Zoom_State.ZoomSelect
                    # Draw rectangle
                    ScopeChShift.zoom_ax.add_patch(ScopeChShift.zoom_rect)
                    # Log flow control
                    logging.debug(f"Start zoom at x: {ScopeChShift.zoom_start_x} y:{ScopeChShift.zoom_start_y}.")
                    # Update plot
                    plt.draw()
                else:
                    # Log flow control
                    logging.debug(f"Button no. {event.button} is pressed when zoom state is {ScopeChShift.zoom_state}.")

            else:
                # Log flow control
                logging.debug(f"Button no. {event.button} is pressed inside the plot.")

        else:
            # Log flow control
            logging.debug("Button press outside of the plot.")

    @staticmethod
    def __on_motion(event: any):
        """
        Provide the current mouse position in case of mouse movement.

        This callback method is private (only for internal usage)
        Called by matplotlib-Event on button press

        :param event: Container with some information: x and y position in pixel, x and y position in data scale,...
        :type event: any
        """
        # Check Zoomstate
        if ScopeChShift.zoom_state == ScopeChShift.Zoom_State.ZoomSelect:
            # Check, if movement is within area and button still pressed
            if event.button == 1:
                if ScopeChShift.zoom_rect is not None and event.inaxes and ScopeChShift.shiftfigbox.contains(event.x, event.y):
                    # Overtake zoom end values
                    width = event.xdata - ScopeChShift.zoom_start_x
                    height = event.ydata - ScopeChShift.zoom_start_y
                    ScopeChShift.zoom_rect.set_width(width)
                    ScopeChShift.zoom_rect.set_height(height)
                    ScopeChShift.zoom_rect.set_xy((ScopeChShift.zoom_start_x, ScopeChShift.zoom_start_y))
                    # Update plot
                    plt.draw()
            else:
                # Remove rectangle and reset zoom state
                ScopeChShift.zoom_rect.remove()
                ScopeChShift.zoom_rect = None
                # Change Zoomstate to NoZoom
                ScopeChShift.zoom_state = ScopeChShift.Zoom_State.NoZoom
                # Update plot
                plt.draw()
                # Log flow control
                logging.debug(f"Right mouse button is pressed while zoom selection at: y={event.xdata} y={event.ydata}.")

    @staticmethod
    def __on_release(event: any):
        """
        Notify the mouse button release.

        This callback method is private (only for internal usage)
        Called by matplotlib-Event on button press

        :param event: Container with some information: x and y position in pixel, x and y position in data scale,...
        :type event: any
        """
        # Check, if the left button was released
        if event.button == 1:
            # Check Zoomstate
            if ScopeChShift.zoom_state == ScopeChShift.Zoom_State.ZoomSelect:
                # Check, if the button was pressed within the area
                if ScopeChShift.zoom_rect is not None and event.inaxes and ScopeChShift.shiftfigbox.contains(event.x, event.y):
                    # Define the area for zoom
                    ScopeChShift.zoom_end_x, ScopeChShift.zoom_end_y = event.xdata, event.ydata
                    # Sort data
                    if ScopeChShift.zoom_start_x > event.xdata:
                        ScopeChShift.zoom_end_x = ScopeChShift.zoom_start_x
                        ScopeChShift.zoom_start_x = event.xdata
                    if ScopeChShift.zoom_start_y > event.ydata:
                        ScopeChShift.zoom_end_y = ScopeChShift.zoom_start_y
                        ScopeChShift.zoom_start_y = event.ydata
                    # Check minimum zoom limits of x-axe
                    cur_zoom_delta = ScopeChShift.zoom_end_x - ScopeChShift.zoom_start_x
                    # Read  range
                    cur_start_x, cur_end_x = ScopeChShift.zoom_ax.get_xlim()
                    # Check if the minimum zoom is reached
                    if cur_end_x - cur_start_x <= ScopeChShift.zoom_delta_x:
                        # Overtake rect
                        ScopeChShift.zoom_end_x = cur_end_x
                        ScopeChShift.zoom_start_x = cur_start_x
                        # Log flow control
                        logging.debug("minimum zoom in x-direction is reached. Zooming in x-direction is denied.")
                    elif cur_zoom_delta < ScopeChShift.zoom_delta_x:
                        logging.debug(f"Selected range in x-direction {ScopeChShift.zoom_start_x} to {event.xdata} is too small.")
                        ScopeChShift.zoom_end_x = ScopeChShift.zoom_end_x + (ScopeChShift.zoom_delta_x - cur_zoom_delta) / 2
                        # Check, if maximum limit is exceed
                        if ScopeChShift.zoom_end_x > ScopeChShift.display_max_x:
                            ScopeChShift.zoom_end_x = ScopeChShift.display_max_x
                            ScopeChShift.zoom_start_x = ScopeChShift.zoom_end_x - ScopeChShift.zoom_delta_x
                        else:  # Calculate minimum window value
                            ScopeChShift.zoom_start_x = ScopeChShift.zoom_start_x - (ScopeChShift.zoom_delta_x - cur_zoom_delta) / 2
                        # Check, if minimum limit is exceed
                        if ScopeChShift.zoom_start_x < ScopeChShift.display_min_x:
                            ScopeChShift.zoom_start_x = ScopeChShift.display_min_x
                            ScopeChShift.zoom_end_x = ScopeChShift.zoom_end_x + ScopeChShift.zoom_delta_x
                        # Log flow control
                        logging.debug(f"Range in x-direction is corrected to {ScopeChShift.zoom_start_x} to {ScopeChShift.zoom_end_x}.")
                    # Check minimum zoom limits of y-axe
                    cur_zoom_delta = ScopeChShift.zoom_end_y - ScopeChShift.zoom_start_y
                    # Read  range
                    cur_start_y, cur_end_y = ScopeChShift.zoom_ax.get_ylim()
                    # Check if the minimum zoom is reached
                    if cur_end_y - cur_start_y <= ScopeChShift.zoom_delta_y:
                        # Overtake rect
                        ScopeChShift.zoom_end_y = cur_end_y
                        ScopeChShift.zoom_start_y = cur_start_y
                        # Log flow control
                        logging.debug("minimum zoom in y-direction is reached. Zooming in y-direction is denied.")
                    elif cur_zoom_delta < ScopeChShift.zoom_delta_y:
                        logging.debug(f"Selected range in y-direction {ScopeChShift.zoom_start_y} to {event.ydata} is too small.")
                        ScopeChShift.zoom_end_y = ScopeChShift.zoom_end_y + (ScopeChShift.zoom_delta_y - cur_zoom_delta) / 2
                        # Check, if maximum limit is exceed
                        if ScopeChShift.zoom_end_y > ScopeChShift.display_max_y:
                            ScopeChShift.zoom_end_y = ScopeChShift.display_max_y
                            ScopeChShift.zoom_start_y = ScopeChShift.zoom_end_y - ScopeChShift.zoom_delta_y
                        else:  # Calculate minimum window value
                            ScopeChShift.zoom_start_y = (
                                ScopeChShift.zoom_start_y - (ScopeChShift.zoom_delta_y - cur_zoom_delta) / 2)
                        # Check, if minimum limit is exceed
                        if ScopeChShift.zoom_start_y < ScopeChShift.display_min_y:
                            ScopeChShift.zoom_start_y = ScopeChShift.display_min_y
                            ScopeChShift.zoom_end_y = ScopeChShift.zoom_end_y + ScopeChShift.zoom_delta_y
                            # Log flow control
                            logging.debug(f"Range in y-direction is corrected to {ScopeChShift.zoom_start_y} to {ScopeChShift.zoom_end_y}.")
                    width = ScopeChShift.zoom_end_x - ScopeChShift.zoom_start_x
                    height = ScopeChShift.zoom_end_y - ScopeChShift.zoom_start_y
                    ScopeChShift.zoom_rect.set_width(width)
                    ScopeChShift.zoom_rect.set_height(height)
                    ScopeChShift.zoom_rect.set_xy((ScopeChShift.zoom_start_x, ScopeChShift.zoom_start_y))
                    # Update zoom state
                    ScopeChShift.zoom_state = ScopeChShift.Zoom_State.ZoomConfirm
                    # Update plot
                    plt.draw()
                elif ScopeChShift.zoom_rect is not None:
                    # Remove rectangle and reset zoom state
                    ScopeChShift.zoom_rect.remove()
                    ScopeChShift.zoom_rect = None
                    ScopeChShift.zoom_state = ScopeChShift.Zoom_State.NoZoom
                    # Update plot
                    plt.draw()
                    # Log flow control
                    logging.debug("Rectangle removed and reset zoom state to 'NoZoom'.")
            # Check if Zoomstate is ZoomConfirm
            elif ScopeChShift.zoom_state == ScopeChShift.Zoom_State.ZoomConfirm:
                # Check, if the button was pressed within the area
                if ScopeChShift.zoom_rect is not None and event.inaxes and ScopeChShift.shiftfigbox.contains(event.x, event.y):
                    if (event.xdata > ScopeChShift.zoom_start_x) and (event.ydata > ScopeChShift.zoom_start_y) and \
                       (event.xdata < ScopeChShift.zoom_end_x) and (event.ydata < ScopeChShift.zoom_end_y):
                        # Zoom into area
                        ScopeChShift.zoom_ax.set_xlim(min(ScopeChShift.zoom_start_x, ScopeChShift.zoom_end_x),
                                                      max(ScopeChShift.zoom_start_x, ScopeChShift.zoom_end_x))
                        ScopeChShift.zoom_ax.set_ylim(min(ScopeChShift.zoom_start_y, ScopeChShift.zoom_end_y),
                                                      max(ScopeChShift.zoom_start_y, ScopeChShift.zoom_end_y))
                # Remove rectangle and reset zoom state
                ScopeChShift.zoom_rect.remove()
                ScopeChShift.zoom_rect = None
                ScopeChShift.zoom_state = ScopeChShift.Zoom_State.NoZoom
                # Update plot
                plt.draw()
                # Log flow control
                logging.debug(f"Confirmed with left mouse button release at x,y: {event.xdata},{event.ydata}.")
        # Check if left button was released and Zoomstate was NoZoom and it was within area
        elif (event.button == 3 and event.inaxes and ScopeChShift.shiftfigbox.contains(event.x, event.y) \
              and ScopeChShift.zoom_state is not ScopeChShift.Zoom_State.ZoomSelect):
            # Check, if zoom state is conform
            if ScopeChShift.zoom_state == ScopeChShift.Zoom_State.ZoomConfirm:
                # Remove rectangle and reset zoom state
                ScopeChShift.zoom_rect.remove()
                ScopeChShift.zoom_rect = None
                ScopeChShift.zoom_state = ScopeChShift.Zoom_State.NoZoom
                # Log flow control
                logging.debug(f"Zoom range is not confirmed with right mouse button at: y={event.xdata} y={event.ydata}.")
            else:
                # Zoom out: Get  zoom
                cur_y_start, cur_y_end = ScopeChShift.zoom_ax.get_ylim()
                cur_x_start, cur_x_end = ScopeChShift.zoom_ax.get_xlim()
                # Calculate factor 2 for y dimension
                delta_half = (cur_y_end - cur_y_start) / 2
                cur_y_start = cur_y_start - delta_half
                cur_y_end = cur_y_end + delta_half
                # Check maximal limits
                if cur_y_start < ScopeChShift.display_min_y:
                    cur_y_start = ScopeChShift.display_min_y
                if cur_y_end > ScopeChShift.display_max_y:
                    cur_y_end = ScopeChShift.display_max_y
                # Calculate factor 2 for x dimension
                delta_half = (cur_x_end - cur_x_start) / 2
                cur_x_start = cur_x_start - delta_half
                cur_x_end = cur_x_end + delta_half
                # Check maximal limits
                if cur_x_start < ScopeChShift.display_min_x:
                    cur_x_start = ScopeChShift.display_min_x
                if cur_x_end > ScopeChShift.display_max_x:
                    cur_x_end = ScopeChShift.display_max_x
                # Zoom out
                ScopeChShift.zoom_ax.set_xlim(cur_x_start, cur_x_end)
                ScopeChShift.zoom_ax.set_ylim(cur_y_start, cur_y_end)
                # Log flow control
                logging.debug(
                    f"Zoom out to x-range: {cur_x_start} to {cur_x_end} and y-range:{cur_y_start} to {cur_y_end}.")
            # Update plot
            plt.draw()
            logging.debug(f"Confirmed with right mouse button release at x,y: {event.xdata},{event.ydata}.")



"""Example file to demonstrate the scope functionality."""
import pysignalscope as pss

# Use this line to change the fonts into LaTeX font.
# pss.global_plot_settings_font_latex()

# ------------------------------------------
# Example 1: Read curves from tektronix scope csv file,
#            shift 2 curves
#            plot the signals and perform FFT
# ------------------------------------------

# Read curves from scope csv file
[voltage_prim, voltage_sec, current_prim, current_sec] = pss.Scope.from_tektronix('scope_example_data_tek.csv')

# Add labels and units to channels
voltage_prim = pss.Scope.modify(voltage_prim, label='voltage primary', unit='V')
voltage_sec = pss.Scope.modify(voltage_sec, label='voltage secondary', unit='V')
current_prim = pss.Scope.modify(current_prim, label='current primary', unit='A')
current_sec = pss.Scope.modify(current_sec, label='current secondary', unit='A')

# Gain and DC-offset for the secondary current
current_sec = pss.Scope.modify(current_sec, data_factor=1.3, data_offset=10)

# Plot channels
fig1 = pss.Scope.plot_channels([voltage_prim, voltage_sec], [current_prim, current_sec], timebase='us')
pss.save_figure(fig1, 'scope_fig.pdf')

# Shift first two channels and plot the shift of these channels
shiftlist = pss.Scope.plot_shiftchannels([voltage_prim, voltage_sec])

# print the list
ch_i = 0
for ch_shift in shiftlist:
    print(f"shifts of channel {ch_i}=x->{ch_shift[0]},y->{ch_shift[1]}")
    ch_i = ch_i+1

# short channels to a single period, perform FFT for current waveforms
current_prim = pss.Scope.short_to_period(current_prim, f0=200000)
current_prim = pss.Scope.modify(current_prim, time_shift=5e-6)
pss.Scope.fft(current_prim)

# ------------------------------------------
# Example 2: Read curves from LeCroy csv-Files and GeckoCIRCUITS. Compare these signals.
# ------------------------------------------

meas_v_ob, meas_il_ib, meas_il_ob = pss.Scope.from_lecroy('scope_example_data_lecroy_1.csv', 'scope_example_data_lecroy_2.csv',
                                                          'scope_example_data_lecroy_3.csv')

meas_v_ob = pss.Scope.modify(meas_v_ob, label='vl_ob_meas', unit='V')
meas_il_ib = pss.Scope.modify(meas_il_ib, label='il_ib_meas', unit='A', color='b')
meas_il_ob = pss.Scope.modify(meas_il_ob, label='il_ob_meas', unit='A', color='m', data_offset=1.5)

# read dataset from gecko simulation
gecko_data = pss.Scope.from_geckocircuits('scope_example_data_gecko', f0=200000)

gecko_il_ib = gecko_data[0]
gecko_il_ib = pss.Scope.modify(gecko_il_ib, label='il_ib_gecko', unit='A', color='r', linestyle="--")
gecko_il_ob = gecko_data[-1]
gecko_il_ob = pss.Scope.modify(gecko_il_ob, data_factor=-1, label='il_ob_gecko', unit='A', color='g')

# compare both waveforms
pss.Scope.compare_channels(meas_il_ib, gecko_il_ib, meas_il_ob, gecko_il_ob, shift=[-67.53e-6, 0, -67.53e-6, 0], timebase='us')

# Shift data
meas_il_ob = pss.Scope.modify(meas_il_ob, time_shift=-67.53e-6)
# Calculate sum
sum_il_ob = pss.Scope.add(meas_il_ob, meas_il_ob)

# Calculate difference
diff_il_ob = pss.Scope.subtract(sum_il_ob, meas_il_ob)

# Copy values to new variable
abs_diff_il_ob = pss.Scope.copy(diff_il_ob)
# Calculate absolute values
pss.Scope.calc_abs(abs_diff_il_ob)
# Copy values to new variable
sqr_diff_il_ob = pss.Scope.copy(diff_il_ob)
# Calculate square
pss.Scope.square(sqr_diff_il_ob)
# Calculate values:
# Root means square
rms_diff_il_ob = pss.Scope.calc_rms(diff_il_ob)
# Average value
mean_diff_il_ob = pss.Scope.calc_mean(diff_il_ob)
# Average of absolute values
absmean_diff_il_ob = pss.Scope.calc_absmean(diff_il_ob)
# Print calculated values
print(f"Root mean square = {rms_diff_il_ob}\n"
      f"Average value = {mean_diff_il_ob}\n"
      f"Absolute average value={absmean_diff_il_ob}\n")
# Plot results
pss.Scope.plot_channels([diff_il_ob])
# Plot results
pss.Scope.plot_channels([diff_il_ob, abs_diff_il_ob, sqr_diff_il_ob])



"""Definition of the impedance dataclass."""

# python libraries
import dataclasses
from typing import Union, Optional

# 3rd party libraries
import numpy as np

@dataclasses.dataclass(eq=False)
class ImpedanceChannel:
    """Dataclass for ImpedanceChannel objects in a special format, to keep labels, units and voltages belonging to a certain curve."""

    # mandatory data
    frequency: np.array  #: mandatory: frequency data (mandatory)
    impedance: np.array  #: mandatory: impedance data (mandatory)
    phase_deg: np.array  #: mandatory: phase data in degree (mandatory)

    # optional data
    label: Optional[str]  #: channel label displayed in a plot (optional)
    unit: Optional[str]  #: channel unit displayed in a plot (optional)
    color: Union[str, tuple, None]  #: channel color displayed in a plot (optional)
    linestyle: Optional[str]  #: channel linestyle displayed in a plot (optional)

    # meta data
    source: Optional[str]  #: channel source, additional meta data (optional)

    def __eq__(self, other):
        """Compare two Channel objects."""
        def array_eq(arr1, arr2):
            return (isinstance(arr1, np.ndarray) and isinstance(arr2, np.ndarray) and \
                    arr1.shape == arr2.shape and (arr1 == arr2).all())

        if not isinstance(other, ImpedanceChannel):
            return NotImplemented("Type Channel must be compared to type Channel.")
        return (array_eq(self.frequency, other.frequency) and \
                array_eq(self.impedance, other.impedance) and \
                array_eq(self.phase_deg, other.phase_deg) and \
                (self.label == other.label) and \
                (self.unit == other.unit) and \
                (self.color == other.color) and \
                (self.source == other.source) and \
                (self.linestyle == other.linestyle))



"""Unit tests for the scope module."""

# python libraries
import pytest
import os

# 3rd party libraries
import numpy as np
import numpy.testing

# key must be set before import of pysignalscope. Disables the GUI inside the test machine.
os.environ["IS_TEST"] = "True"

# own libraries
import pysignalscope as pss

#########################################################################################################
# test of generate_channel
#########################################################################################################

# parameterset for values
@pytest.mark.parametrize("tst_channel_time,tst_channel_data,exp_scope_obj_or_err,error_flag", [
    # --invalid inputs----
    # no input or missing input channel_time and channel_data - raise type error
    (None, None, TypeError, True),
    ([1, 2, 3], None, TypeError, True),
    (None, [1, 2, 3], TypeError, True),
    # empty lists
    ([], [], ValueError, True),
    # different length of time and data must raise value error
    ([1, 2, 3], [1, 2], ValueError, True),
    # invalid time data positive values
    ([3, 2, 1], [1, 2, 3], ValueError, True),
    # invalid time data negative values. Time values in wrong order.
    ([-1, -2, -3], [1, 2, 3], ValueError, True),
    # invalid time data negative and positive values. Time values in wrong order.
    ([-1, -3, 1], [1, 2, 3], ValueError, True),
    # invalid time value
    ([np.nan, 2, 3], [0, 2, -3.2], ValueError, True),
    # invalid data value not defined, minus infinite, plus infinite
    ([1, 2, 3], [-np.nan, 2, -3.2], ValueError, True),
    ([1, 2, 3], [-np.inf, 2, -3.2], ValueError, True),
    ([1, 2, 3], [np.inf, 2, -3.2], ValueError, True),
    # --valid inputs----
    # channel_time: non-equidistant values and negative valid values
    ([-3.3, -2.2, -1.1, 0, 1.2], [-1, -2.1, -3.2, 4.4, -2.7], {"channel_time": [-3.3, -2.2, -1.1, 0, 1.2], "channel_data": [-1, -2.1, -3.2, 4.4, -2.7]}, False),
    # valid positive data, mixed int and float
    ([1, 2, 3], [1, 2, 3.1], {"channel_time": [1, 2, 3], "channel_data": [1, 2, 3.1]}, False),
    # valid negative data, mixed int and float
    ([1, 2, 3], [-1, -2.1, -3.2], {"channel_time": [1, 2, 3], "channel_data": [-1, -2.1, -3.2]}, False),
    # very high, very low and very small mixed values
    ([1, 2, 3], [1e25, -3.4e34, 3.1e-17], {"channel_time": [1, 2, 3], "channel_data": [1e25, -3.4e34, 3.1e-17]}, False)
])
# definition of the testfunction
def test_values_of_generate_channel(tst_channel_time: [], tst_channel_data: [], exp_scope_obj_or_err, error_flag: bool):
    """Test generate_channel() method according values.

    :param tst_channel_time: test time series
    :type tst_channel_time: Union[List[float], np.ndarray]
    :param tst_channel_data: test data series
    :type tst_channel_data: Union[List[float], np.ndarray]
    :param exp_scope_obj_or_err: expected channel time data and value data or expected error
    :type exp_scope_obj_or_err: any
    :param error_flag: flag to indicate, if an error or a valid result is expected
    :type error_flag: bool

    """
    # Check if expected test result is no error
    if error_flag is False:
        # channel_time: non-equidistant values and negative valid values
        scope_object = pss.Scope.generate_channel(time=tst_channel_time, data=tst_channel_data)
        numpy.testing.assert_equal(scope_object.time, exp_scope_obj_or_err["channel_time"])
        numpy.testing.assert_equal(scope_object.data, exp_scope_obj_or_err["channel_data"])
    else:  # generate_channel raises an error
        with pytest.raises(exp_scope_obj_or_err):
            pss.Scope.generate_channel(tst_channel_time, tst_channel_data)

# parameterset for attributes
@pytest.mark.parametrize(
    "tst_channel_label,tst_channel_unit,tst_channel_color,tst_channel_source,"
    "tst_channel_linestyle,exp_error, error_flag", [
        # --invalid inputs----
        # wrong type for one attribute
        (100.1, "A", None, "scope 11", None, TypeError, True),
        ("Label", 1, None, "scope 11", None, TypeError, True),
        ("Label", "Unit", 1.2, "scope 11", None, TypeError, True),
        ("Label", "Unit", "Color", 4, None, TypeError, True),
        ("Label", "Unit", "Color", "scope 11", 18, TypeError, True),
        # --valid inputs----
        # all inputs are set
        ("test 1", "A", "red", "scope 11", "--", None, False),
        # some input is set
        (None, "A", None, "scope 11", None, None, False),
        # no input is set
        (None, None, None, None, None, None, False)
    ])
# definition of the testfunction
def test_attributes_of_test_generate_channel(tst_channel_label, tst_channel_unit, tst_channel_color,
                                             tst_channel_source, tst_channel_linestyle, exp_error, error_flag: bool):
    """Test generate_channel() method according attributes.

    :param tst_channel_label: test variable for channel label
    :type tst_channel_label: any
    :param tst_channel_unit: test variable for channel unit
    :type tst_channel_unit: any
    :param tst_channel_color: test variable for channel color
    :type tst_channel_color: any
    :param tst_channel_source: test variable for channel source
    :type tst_channel_source: any
    :param tst_channel_linestyle: test variable for channel linestyle
    :type tst_channel_linestyle: any
    :param exp_error: expected error
    :type exp_error: any
    :param error_flag: flag to indicate, if an error or a valid result is expected
    :type error_flag: bool
    """
    # Define value input
    tst_channel_time = [-3.3, -2.2, -1.1, 0, 1.2]
    tst_channel_data = [-1, -2.1, -3.2, 4.4, -2.7]
    # Check if expected test result is no error
    if error_flag is False:
        # channel_time: non-equidistant values and negative valid values
        scope_object = pss.Scope.generate_channel(time=tst_channel_time,
                                                  data=tst_channel_data,
                                                  label=tst_channel_label,
                                                  unit=tst_channel_unit,
                                                  color=tst_channel_color,
                                                  source=tst_channel_source,
                                                  linestyle=tst_channel_linestyle)
        # verification of function result
        numpy.testing.assert_equal(scope_object.time, tst_channel_time)
        numpy.testing.assert_equal(scope_object.data, tst_channel_data)
        numpy.testing.assert_equal(scope_object.label, tst_channel_label)
        numpy.testing.assert_equal(scope_object.unit, tst_channel_unit)
        numpy.testing.assert_equal(scope_object.color, tst_channel_color)
        numpy.testing.assert_equal(scope_object.source, tst_channel_source)
        numpy.testing.assert_equal(scope_object.linestyle, tst_channel_linestyle)
    else:  # generate_channel raises an error
        with pytest.raises(exp_error):
            scope_object = pss.Scope.generate_channel(time=tst_channel_time,
                                                      data=tst_channel_data,
                                                      label=tst_channel_label,
                                                      unit=tst_channel_unit,
                                                      color=tst_channel_color,
                                                      source=tst_channel_source,
                                                      linestyle=tst_channel_linestyle)

#########################################################################################################
# test of from_numpy
#########################################################################################################


# test frequency
frequency1 = 200.6
frequency2 = 20000

# parameterset for values
@pytest.mark.parametrize("valid_input_flag, tst_mode, tst_f0, exp_factor_or_error, error_flag", [
    # --invalid inputs----
    # wrong type for one attribute
    (True, "wrong_mode", None, ValueError, True),
    (True, 2, None, ValueError, True),
    (True, None, "Freq", ValueError, True),
    (True, "rad", None, ValueError, True),
    (True, None, None, ValueError, True),
    (True, None, frequency1, ValueError, True),
    (False, "time", None, ValueError, True),
    # --valid inputs----
    (True, "time", None, 1, False),
    (True, "time", frequency1, 1, False),
    (True, "rad", frequency1, (1 / 2 / np.pi / frequency1), False),
    (True, "rad", frequency2, (1 / 2 / np.pi / frequency2), False),
    (True, "deg", frequency1, (1 / 360 / frequency1), False),
    (True, "deg", frequency2, (1 / 360 / frequency2), False)
])
# definition of the testfunction
def test_from_numpy(valid_input_flag: bool, tst_mode, tst_f0, exp_factor_or_error, error_flag: bool):
    """Test for the method from_numpy().

    :param valid_input_flag: flag to indicate that valid time and value data are to use
    :type valid_input_flag: bool
    :param tst_mode: test variable for mode
    :type tst_mode: any
    :param tst_f0: test variable for the frequency
    :type tst_f0: any
    :param exp_factor_or_error: expected factor for the result vector or expected error
    :type exp_factor_or_error: any
    :param error_flag: flag to indicate, if an error or a valid result is expected
    :type error_flag: bool
    """
    # Define value input
    tst_time = [1.1, 2.2, 3.3, 4]
    tst_time_invalid = [1.1, -1.1, 3.3, 4]
    tst_data = [-1, -2.1, -3.2, 4.4]

    tst_label = "DataLabel"
    tst_unit = "UnitOfData"

    # valid vector
    period_vector_t_i = np.array([tst_time, tst_data])
    # invalid vector
    period_vector_t_i_invalid = np.array([tst_time_invalid, tst_data])

    # Check if expected test result is no error
    if error_flag is False:
        # channel_time: non-equidistant values and negative valid values
        scope_object = pss.Scope.from_numpy(period_vector_t_i, mode=tst_mode, f0=tst_f0,
                                            label=tst_label, unit=tst_unit)
        # verification of attributes
        numpy.testing.assert_equal(scope_object.label, tst_label)
        numpy.testing.assert_equal(scope_object.unit, tst_unit)
        # verification of function result
        np.testing.assert_array_almost_equal(scope_object.time, np.array(tst_time) * exp_factor_or_error)
        np.testing.assert_array_equal(scope_object.data, tst_data)
    else:  # generate_channel raises an error
        if valid_input_flag is True:
            with pytest.raises(exp_factor_or_error):
                scope_object = pss.Scope.from_numpy(period_vector_t_i, mode=tst_mode, f0=tst_f0,
                                                    label=tst_label, unit=tst_unit)
        else:
            with pytest.raises(exp_factor_or_error):
                scope_object = pss.Scope.from_numpy(period_vector_t_i_invalid, mode=tst_mode, f0=tst_f0,
                                                    label=tst_label, unit=tst_unit)

#########################################################################################################
# test of low_pass_filter
#########################################################################################################


# channel values
test_vec1 = pss.Scope.generate_channel([0, 1, 2, 3, 4, 5, 6], [1, 4, 2, 3, 7, 3, 2])

# result values
test_res1 = [0.99927604, 2.26610791, 2.85423117, 3.5885494, 4.09641649, 3.33691443, 1.99801723]
test_res2 = [0.7568143, 0.98001724, 1.15909406, 1.2985092, 1.37680805, 1.36477982, 1.29328745]

# parameterset for values
@pytest.mark.parametrize("tst_vector, tst_order, tst_angular_freq, exp_result_or_error, error_flag", [
    # --invalid inputs----
    # wrong filter order type
    (test_vec1, 1.4, 0.5, TypeError, True),
    # negative filter order
    (test_vec1, -3, 0.5, ValueError, True),
    # wrong filter frequency type
    (test_vec1, 1, "Freq", TypeError, True),
    # wrong frequency value
    (test_vec1, 1, 1.4, ValueError, True),
    (test_vec1, 1, -2.2, ValueError, True),
    # insert not a scope type
    (5, 1, 0.3, TypeError, True),
    # --valid inputs----
    (test_vec1, 1, 0.3, test_res1, False),
    (test_vec1, None, None, test_res2, False),
])
# definition of the testfunction
def test_low_pass_filter(tst_vector, tst_order, tst_angular_freq, exp_result_or_error, error_flag: bool):
    """Unit test for low_pass_filter().

    :param tst_vector: test vector corresponds to scope object
    :type tst_vector: any
    :param tst_order: test variable for mode order or None indicates the default value test
    :type tst_order: any
    :param tst_angular_freq: test variable for the frequency angular_freq
    :type tst_angular_freq: any
    :param exp_result_or_error: expected test result vector or expected error
    :type exp_result_or_error: any
    :param error_flag: flag to indicate, if an error or a valid result is expected
    :type error_flag: bool
    """
    # Check if expected test result is no error
    if error_flag is False:
        # test default value test
        if isinstance(tst_order, (float, int)):
            # test with valid values
            filter_current_prim_1 = pss.Scope.low_pass_filter(tst_vector, tst_order, tst_angular_freq)
        else:
            # test of function default values
            filter_current_prim_1 = pss.Scope.low_pass_filter(tst_vector)

        # verification of function result
        numpy.testing.assert_array_almost_equal(exp_result_or_error, filter_current_prim_1.data)
    else:  # generate_channel raises an error
        with pytest.raises(exp_result_or_error):
            filter_current_prim_1 = pss.Scope.low_pass_filter(
                tst_vector, order=tst_order, angular_frequency_rad=tst_angular_freq)

#########################################################################################################
# test of derivative
#########################################################################################################


# channel values
test_vec1 = pss.Scope.generate_channel([0, 1, 2, 3, 4, 5, 6], [1, 4, 2, 3, 7, 3, 2])

# result values
test_res1 = [6, 1, 0, 2, 0, -2, 0]

# parameterset for values
@pytest.mark.parametrize("tst_vector, tst_order, exp_result_or_error, error_flag", [
    # --invalid inputs----
    # wrong order type
    (test_vec1, 3.3, TypeError, True),
    # negative order
    (test_vec1, -3, ValueError, True),
    # insert not a scope type
    (5, 1, TypeError, True),
    # --valid inputs----
    (test_vec1, 1, test_res1, False),
    (test_vec1, None, test_res1, False),
])
# definition of the testfunction
def test_derivative(tst_vector, tst_order, exp_result_or_error, error_flag: bool):
    """Test the derivative method.

    :param tst_vector: test vector corresponds to scope object
    :type tst_vector: any
    :param tst_order: test variable for mode order or None indicates the default value test
    :type tst_order: any
    :param exp_result_or_error: expected test result vector or expected error
    :type exp_result_or_error: any
    :param error_flag: flag to indicate, if an error or a valid result is expected
    :type error_flag: bool
    """
    # Check if expected test result is no error
    if error_flag is False:
        # test default value test
        if isinstance(tst_order, (float, int)):
            # test with valid values
            sample_scope_object_1st_derivative = pss.Scope.derivative(tst_vector, tst_order)
        else:
            # test of function default values
            sample_scope_object_1st_derivative = pss.Scope.derivative(tst_vector)

        # verification of function result
        numpy.testing.assert_array_equal(exp_result_or_error, sample_scope_object_1st_derivative.data)
    else:  # generate_channel raises an error
        with pytest.raises(exp_result_or_error):
            sample_scope_object_1st_derivative = pss.Scope.derivative(tst_vector, tst_order)

#########################################################################################################
# test of unify_sampling_rate
#########################################################################################################


# definition of the scope objects for test vector and result vector
ch1 = pss.Scope.generate_channel([-1, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
                                 [-8, -2, 10.5, 11, 13, 14, 16, 20, 17, 14, 9, 1])
ch2 = pss.Scope.generate_channel([-0.5, 0, 0.5, 1, 1.5, 2, 2.5, 3, 3.5, 4, 4.5],
                                 [10, 2, 7.5, -2.5, 4, 8, 4, 10, 2, 20, 5])
ch3 = pss.Scope.generate_channel([13, 14, 15, 16, 17, 18, 19, 21],
                                 [13, 14, 16, 20, 17, 14, 9, 1])
ch4 = pss.Scope.generate_channel([10], [1.4])

# min mastermode is false
uch11 = pss.Scope.generate_channel([-1.0, -0.5, 0.0, 0.5, 1.0, 1.5, 2.0, 2.5, 3.0, 3.5, 4.0, 4.5, 5.0, 5.5,
                                    6.0, 6.5, 7.0, 7.5, 8.0, 8.5, 9.0, 9.5, 10.0],
                                   [-8.0, -5.0, -2.0, 4.25, 10.5, 10.75, 11.0, 12.0, 13.0, 13.5, 14.0,
                                    15.0, 16.0, 18.0, 20.0, 18.5, 17.0, 15.5, 14.0, 11.5, 9.0, 5.0, 1.0])
uch12 = pss.Scope.generate_channel([-0.5, 0.0, 0.5, 1.0, 1.5, 2.0, 2.5, 3.0, 3.5, 4.0, 4.5],
                                   [10.0, 2.0, 7.5, -2.5, 4.0, 8.0, 4.0, 10.0, 2.0, 20.0, 5.0])

uch13 = pss.Scope.generate_channel([13.0, 13.5, 14.0, 14.5, 15.0, 15.5, 16.0, 16.5, 17.0, 17.5, 18.0,
                                    18.5, 19.0, 19.5, 20.0, 20.5, 21],
                                   [13.0, 13.5, 14.0, 15.0, 16.0, 18.0, 20.0, 18.5, 17.0, 15.5, 14.0, 11.5,
                                    9.0, 7.0, 5.0, 3.0, 1.0])

# max mastermode is false
uch21 = pss.Scope.generate_channel([-1, 1, 3, 5, 7, 9], [-8, 10.5, 13, 16, 17, 9])
uch22 = pss.Scope.generate_channel([1.0, 3.0], [-2.5, 10.0])
uch23 = pss.Scope.generate_channel([13, 15, 17, 19, 21], [13, 16, 17, 9, 1])

# avg mastermode is true
uch31 = pss.Scope.generate_channel([-1, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
                                   [-8, -2, 10.5, 11, 13, 14, 16, 20, 17, 14, 9, 1])

uch32 = pss.Scope.generate_channel([0, 1, 2, 3, 4], [2, -2.5, 8, 10, 20])
uch33 = pss.Scope.generate_channel([13, 14, 15, 16, 17, 18, 19, 20, 21],
                                   [13, 14, 16, 20, 17, 14, 9, 5.0, 1])

# avg + shift= 0.5 mastermode is true
uch41 = pss.Scope.generate_channel([-0.5, 0.5, 1.5, 2.5, 3.5, 4.5, 5.5, 6.5, 7.5, 8.5, 9.5],
                                   [-5.0, 4.25, 10.75, 12.0, 13.5, 15.0, 18.0, 18.5, 15.5, 11.5, 5.0])
uch42 = pss.Scope.generate_channel([-0.5, 0.5, 1.5, 2.5, 3.5, 4.5], [10.0, 7.5, 4.0, 4.0, 2.0, 5.0])
uch43 = pss.Scope.generate_channel([13.5, 14.5, 15.5, 16.5, 17.5, 18.5, 19.5, 20.5],
                                   [13.5, 15.0, 18.0, 18.5, 15.5, 11.5, 7.0, 3.0])

# user sample rate=1/3  shift= -0.5 mastermode is true
uch51 = pss.Scope.generate_channel([1.5, 4.5, 7.5], [10.75, 15.0, 15.5])
uch52 = pss.Scope.generate_channel([1.5, 4.5], [4., 5.])
uch53 = pss.Scope.generate_channel([13.5, 16.5, 19.5], [13.5, 18.5, 7.0])

# avg mastermode is False shift=-0.5
uch61 = pss.Scope.generate_channel([-0.64285714, 0.21428571, 1.07142857, 1.92857143, 2.78571429, 3.64285714,
                                    4.5, 5.35714286, 6.21428571, 7.07142857, 7.92857143, 8.78571429, 9.64285714],
                                   [-5.85714286, 0.67857143, 10.53571429, 10.96428571, 12.57142857,
                                    13.64285714, 15., 17.42857143, 19.35714286, 16.78571429, 14.21428571, 10.07142857,
                                    3.85714286])
uch62 = pss.Scope.generate_channel([0.21428571, 1.07142857, 1.92857143, 2.78571429, 3.64285714, 4.5],
                                   [4.35714286, -1.57142857, 7.42857143, 7.42857143, 7.14285714, 5.0])
uch63 = pss.Scope.generate_channel([13.07142857, 13.92857143, 14.78571429, 15.64285714, 16.5, 17.35714286,
                                    18.21428571, 19.07142857, 19.92857143, 20.78571429],
                                   [13.07142857, 13.92857143, 15.57142857, 18.57142857, 18.5, 15.92857143,
                                    12.92857143, 8.71428571, 5.28571429, 1.85714286])

# testvectors and resultvector
# valid vector
tst_vector1 = [ch1, ch2, ch3]
# invalid vector
tst_vector2 = [ch1, 2, ch3]
tst_vector3 = [ch1, ch2, ch4]
# resultvectors
# min mastermode is false
result_vec1 = [uch11, uch12, uch13]
# max mastermode is false
result_vec2 = [uch21, uch22, uch23]
# avg mastermode is true
result_vec3 = [uch31, uch32, uch33]
# avg + shift= 0.5 mastermode is true
result_vec4 = [uch41, uch42, uch43]
# user sample rate=1/3  shift= -0.5 mastermode is true
result_vec5 = [uch51, uch52, uch53]
# avg mastermode is False shift=-5.5
result_vec6 = [uch61, uch62, uch63]

# parameterset for values
@pytest.mark.parametrize("tst_vector, tst_sample_calc_mode, tst_sampling_rate, tst_shift, tst_mastermode, exp_scope_obj_or_err, error_flag", [
    # --invalid inputs----
    # no input or missing input channel_time and channel_data - raise type error
    # Wrong  or invalid sample_calc_mode-keyword
    (tst_vector1, "murx", None, 0, False, ValueError, True),
    (tst_vector1, 4, None, 0, False, TypeError, True),
    # invalid channel
    (tst_vector2, "min", None, 0, False, TypeError, True),
    (tst_vector3, "min", None, 0, False, ValueError, True),
    ([], 4, None, "min", False, TypeError, True),
    # invalid or too high or missing sampling rate
    # (tst_vector1, "min", True, 0, False, TypeError, True),
    (tst_vector1, "user", -10, 0, False, ValueError, True),
    (tst_vector1, "user", 1E9, 0, False, ValueError, True),
    (tst_vector1, "user", None, 0, False, TypeError, True),
    # --valid inputs----
    (tst_vector1, "min", None, 0, False, result_vec1, False),
    (tst_vector1, "max", None, 0, False, result_vec2, False),
    (tst_vector1, "avg", None, 0, True, result_vec3, False),
    (tst_vector1, "avg", None, 0.5, True, result_vec4, False),
    (tst_vector1, "user", (1/3), -0.5, True, result_vec5, False),
    (tst_vector1, "avg", None, -0.5, False, result_vec6, False),
    # test default values
    (tst_vector1, "min", None, None, None, result_vec1, False),
])
# testfunction
def test_of_unify_sampling_rate(tst_vector, tst_sample_calc_mode, tst_sampling_rate, tst_shift, tst_mastermode, exp_scope_obj_or_err, error_flag: bool):
    """Test generate_channel() method according values.

    :param tst_vector: test variable for parameter channel_datasets
    :type tst_vector: any
    :param tst_sample_calc_mode: test variable for parameter sample_calc_mode
    :type tst_sample_calc_mode: any
    :param tst_sampling_rate: sampling rate defined by the user (only valid, if granularity is set to 'user'
    :type tst_sampling_rate: any
    :param tst_shift: shift of the sample rate from origin. None corresponds to a shift to first time point of first channel
    :type tst_shift: any
    :param tst_mastermode: Indicate, if only the first data set or all data sets are used for sampling rate calculation
    :type tst_mastermode: any
    :param exp_scope_obj_or_err: expected channel time data and value data or expected error
    :type exp_scope_obj_or_err: any
    :param error_flag: flag to indicate, if an error or a valid result is expected
    :type error_flag: bool

    """
    # Check if expected test result is no error
    if error_flag is False:
        # channel_time: non-equidistant values and negative valid values
        result_list = pss.Scope.unify_sampling_rate(*tst_vector, sample_calc_mode=tst_sample_calc_mode,
                                                    sampling_rate=tst_sampling_rate, shift=tst_shift, mastermode=tst_mastermode)
        for count, result in enumerate(result_list):
            numpy.testing.assert_array_almost_equal(result.time, exp_scope_obj_or_err[count].time)
            numpy.testing.assert_array_almost_equal(result.data, exp_scope_obj_or_err[count].data)
    else:  # generate_channel raises an error
        with pytest.raises(exp_scope_obj_or_err):
            pss.Scope.unify_sampling_rate(*tst_vector, sample_calc_mode=tst_sample_calc_mode,
                                          sampling_rate=tst_sampling_rate, shift=tst_shift, mastermode=tst_mastermode)


#########################################################################################################
# test of eq-operator (alternative stype)
#########################################################################################################

def test_eq():
    """Test __eq__()."""
    ch_1 = pss.Scope.generate_channel([1, 2, 3], [4, 5, 6],
                                      unit="A", label="label", color="red",
                                      source="source", linestyle='--')
    ch_2 = pss.Scope.copy(ch_1)
    # assert both channels are the same
    assert ch_1 == ch_2

    # not the same: different time
    ch_2 = pss.Scope.copy(ch_1)
    ch_2.time = np.array([2, 2.1, 3])
    assert not (ch_1 == ch_2)

    # not the same: different data
    ch_2 = pss.Scope.copy(ch_1)
    ch_2.data = np.array([2, 2.1, 3])
    assert not (ch_1 == ch_2)

    # not the same: different units
    ch_2 = pss.Scope.copy(ch_1)
    ch_2 = pss.Scope.modify(ch_2, unit="U")
    assert not (ch_1 == ch_2)

    # not the same: different labels
    ch_2 = pss.Scope.copy(ch_1)
    ch_2 = pss.Scope.modify(ch_2, label="aaa")
    assert not (ch_1 == ch_2)

    # not the same: different colors
    ch_2 = pss.Scope.copy(ch_1)
    ch_2 = pss.Scope.modify(ch_2, color="blue")
    assert not (ch_1 == ch_2)

    # not the same: different sources
    ch_2 = pss.Scope.copy(ch_1)
    ch_2 = pss.Scope.modify(ch_2, source="asdf")
    assert not (ch_1 == ch_2)

    # not the same: different line styles
    ch_2 = pss.Scope.copy(ch_1)
    ch_2 = pss.Scope.modify(ch_2, label=".-")
    assert not (ch_1 == ch_2)

#########################################################################################################
# test of save_load (alternative stype)
#########################################################################################################

def test_save_load():
    """Unit test for save and load."""
    # assumption: the given scope object is valid
    example = pss.Scope.generate_channel([-1, 2, 3.3], [-4, 5.0, 6.9],
                                         unit="A", label="label", color="red",
                                         source="source", linestyle='--')

    # save + load: working example
    pss.Scope.save(example, "test_example")
    loaded_example = pss.Scope.load("test_example.pkl")
    assert example == loaded_example

    # save: wrong file path type
    with pytest.raises(TypeError):
        pss.Scope.save(example, 123)

    # save: insert wrong scope type
    with pytest.raises(TypeError):
        pss.Scope.save(123, "test_example")

    # load: not a pkl-file
    with pytest.raises(ValueError):
        pss.Scope.load("test_example.m")

    # load: not a string filepath
    with pytest.raises(TypeError):
        pss.Scope.load(123)

    # load: non-existing pkl-file
    with pytest.raises(ValueError):
        pss.Scope.load("test_example_not_existing.pkl")

#########################################################################################################
# test of copy (alternative stype)
#########################################################################################################

def test_copy():
    """Unit test for copy()."""
    # wrong input type
    with pytest.raises(TypeError):
        pss.Scope.copy("not-a-channel")

    # test for valid copy
    object_to_copy = pss.Scope.generate_channel([-1, 2, 3.3], [-4, 5.0, 6.9],
                                                unit="A", label="label", color="red",
                                                source="source", linestyle='--')
    object_copy = pss.Scope.copy(object_to_copy)

    assert object_to_copy == object_copy

#########################################################################################################
# test of add (alternative stype)
#########################################################################################################

def test_add():
    """Unit test for add()."""
    # sample data
    channel_1 = pss.Scope.generate_channel(time=[1, 1.1, 2], data=[1, 2, 3])
    channel_2 = pss.Scope.generate_channel(time=[1, 1.1, 2.2], data=[1, 2, 3])
    channel_3 = pss.Scope.generate_channel(time=[1, 1.1, 2.2], data=[1, 2, 3])
    channel_4 = pss.Scope.generate_channel(time=[1, 1.1, 2.2], data=[2, 4, 6])
    channel_5 = pss.Scope.generate_channel(time=[1, 1.1, 2.2, 3.3], data=[2, 4, 6, 9])

    # wrong input type
    with pytest.raises(TypeError):
        pss.Scope.add(123, "wrong-type")
    # error: different length of channels
    with pytest.raises(ValueError):
        pss.Scope.add(channel_1, channel_5)
    # error: different time steps of channels
    with pytest.raises(ValueError):
        pss.Scope.add(channel_1, channel_2)

    # valid result
    channel_add = pss.Scope.add(channel_2, channel_3)
    np.testing.assert_equal(channel_add.data, channel_4.data)

#########################################################################################################
# test of subtract (alternative stype)
#########################################################################################################

def test_subtract():
    """Unit test for add()."""
    # sample data
    channel_1 = pss.Scope.generate_channel(time=[1, 1.1, 2], data=[1, 2, 3])
    channel_2 = pss.Scope.generate_channel(time=[1, 1.1, 2.2], data=[1, 2, 3])
    channel_3 = pss.Scope.generate_channel(time=[1, 1.1, 2.2], data=[1, 2, 3])
    channel_4 = pss.Scope.generate_channel(time=[1, 1.1, 2.2], data=[0, 0, 0])
    channel_5 = pss.Scope.generate_channel(time=[1, 1.1, 2.2, 3.3], data=[2, 4, 6, 9])

    # wrong input type
    with pytest.raises(TypeError):
        pss.Scope.subtract(123, "wrong-type")
    # error: different length of channels
    with pytest.raises(ValueError):
        pss.Scope.subtract(channel_1, channel_5)
    # error: different time steps of channels
    with pytest.raises(ValueError):
        pss.Scope.subtract(channel_1, channel_2)

    # valid result
    channel_subtract = pss.Scope.subtract(channel_2, channel_3)
    np.testing.assert_equal(channel_subtract.data, channel_4.data)

#########################################################################################################
# test of mean (alternative stype)
#########################################################################################################

def test_mean():
    """Unit test for mean()."""
    # wrong input type
    with pytest.raises(TypeError):
        pss.Scope.calc_mean("not-a-scope-object")

    # mixed signal
    example_channel = pss.Scope.generate_channel([1, 2, 3, 4], [-1, 1, -1, 1])
    rms_example_channel = pss.Scope.calc_mean(example_channel)
    assert rms_example_channel == 0

    # positive signal, negative time steps
    example_channel = pss.Scope.generate_channel([-4.1, -3.1, -2.1, -1.1], [2.5, 1.5, 2.5, 1.5])
    rms_example_channel = pss.Scope.calc_mean(example_channel)
    assert rms_example_channel == 2

    # negative signal
    example_channel = pss.Scope.generate_channel([1, 2, 3, 4], [-2.5, -1.5, -2.5, -1.5])
    rms_example_channel = pss.Scope.calc_mean(example_channel)
    assert rms_example_channel == -2

    # non-equidistant time steps
    example_channel = pss.Scope.generate_channel([-2, -1, 1, 2], [-1, -1, 2, 2])
    rms_example_channel = pss.Scope.calc_mean(example_channel)
    assert rms_example_channel == 0.5

#########################################################################################################
# test of abs_mean (alternative stype)
#########################################################################################################

def test_abs_mean():
    """Unit test for abs_mean()."""
    # wrong input type
    with pytest.raises(TypeError):
        pss.Scope.calc_absmean("not-a-scope-object")

    # mixed signal
    example_channel = pss.Scope.generate_channel([1, 2, 3, 4], [-1, 1, -1, 1])
    rms_example_channel = pss.Scope.calc_absmean(example_channel)
    assert rms_example_channel == 1

    # positive signal, negative time steps
    example_channel = pss.Scope.generate_channel([-4.1, -3.1, -2.1, -1.1], [2.5, 1.5, 2.5, 1.5])
    rms_example_channel = pss.Scope.calc_absmean(example_channel)
    assert rms_example_channel == 2

    # negative signal
    example_channel = pss.Scope.generate_channel([1, 2, 3, 4], [-2.5, -1.5, -2.5, -1.5])
    rms_example_channel = pss.Scope.calc_absmean(example_channel)
    assert rms_example_channel == 2

    # non-equidistant time steps
    example_channel = pss.Scope.generate_channel([-2, -1, 1, 2], [-1, -1, 2, 2])
    rms_example_channel = pss.Scope.calc_absmean(example_channel)
    assert rms_example_channel == 1.5

#########################################################################################################
# test of rms (alternative stype)
#########################################################################################################

def test_rms():
    """Unit test for rms()."""
    # wrong input type
    with pytest.raises(TypeError):
        pss.Scope.calc_rms("not-a-scope-object")

    # mixed signal
    example_channel = pss.Scope.generate_channel([1, 2, 3, 4], [-1, 1, -1, 1])
    rms_example_channel = pss.Scope.calc_rms(example_channel)
    assert rms_example_channel == 1

    # positive signal, negative time steps
    example_channel = pss.Scope.generate_channel([-4.1, -3.1, -2.1, -1.1], [2.5, 1.5, 2.5, 1.5])
    rms_example_channel = pss.Scope.calc_rms(example_channel)
    assert rms_example_channel == 2.0615528128088303

    # negative signal
    example_channel = pss.Scope.generate_channel([1, 2, 3, 4], [-2.5, -1.5, -2.5, -1.5])
    rms_example_channel = pss.Scope.calc_rms(example_channel)
    assert rms_example_channel == 2.0615528128088303

    # non-equidistant time steps
    example_channel = pss.Scope.generate_channel([-2, -1, 1, 2], [-1, -1, 2, 2])
    rms_example_channel = pss.Scope.calc_rms(example_channel)
    assert rms_example_channel == 1.5811388300841898



"""Classes and methods to process scope data (from real scopes or from simulation tools) like in a real scope.

This module set the logging level and provides the function setup_logging to configure
 the logging for each module in a centralized way. It is important to add following line to each
 module so that the module can participate to the centralized logging approach.
 -> from logconfig import setup_logging
 -> import logging
 -> setup_logging()
"""
###################################################################################################
# date         Author  Description
# -----------  ------  ----------------------------------------------------------------------------
# 04.11.2024    ASA    Initial version
###################################################################################################
# - Imports ---------------------------------------------------------------------------------------

import logging
import logging.config

# - Code ------------------------------------------------------------------------------------------

# Set up the logging level for the module
def setup_logging():
    """Set up the logging capability.

    This function call has to be added to each file just after the import block.
    With this file you can modify the logging level.
    The basic logging level corresponds to the logging level, you want to set.
    You can adjust (suppress logging) of signal files by setting a higher level for this particular file.
    """
    # Set logging level for all imported modules
    logging.basicConfig(level=logging.WARNING, format='%(asctime)s - %(name)s - %(levelname)s - %(funcName)s - %(message)s')

    # Set logging level for my modules (only valid if basic config is set less equal selected loglevel)
    # Logger for scope
    Lm_logger = logging.getLogger('scope')
    Lm_logger.setLevel(logging.DEBUG)

    ###############################################################################################
    # Do not change the lines below. These suppress logging information of external modules
    # 'matplotlib' and 'PIL', which are imported by pySignalScope
    ###############################################################################################

    mp_logger = logging.getLogger('matplotlib')
    mp_logger.setLevel(logging.WARNING)

    pil_logger = logging.getLogger('PIL')
    pil_logger.setLevel(logging.WARNING)



"""Functions uses by the scope, e.g. the fft-function."""
from matplotlib import pyplot as plt
from typing import Union, Tuple, Optional, List
import numpy as np

def save_figure(figure: plt.figure, fig_name: str):
    """
    Save the given figure object as pdf.

    :param figure: figure object
    :type figure: matplotlib.pyplot.figure
    :param fig_name: figure name for pdf file naming
    :type fig_name: str
    """
    if isinstance(fig_name, str):
        if fig_name.endswith(".pdf"):
            figure.savefig(f"{fig_name}")
        else:
            figure.savefig(f"{fig_name}.pdf")
    else:
        raise TypeError("figure name must be of type str.")


def fft(period_vector_t_i: Union[List[List[float]], np.ndarray], sample_factor: int = 1000, plot: bool = True, mode: str = 'rad',
        f0: Optional[float] = None, title: str = 'ffT', filter_type: str = 'factor',
        filter_value_factor: float = 0.01, filter_value_harmonic: int = 100,
        figure_size: Optional[Tuple] = None, figure_directory: Optional[str] = None) -> List[List]:
    """
    Calculate the FFT for a given input signal. Input signal is in vector format and should include one period.

    Output vector includes only frequencies with amplitudes > 1% of input signal

    :Minimal Example:

    >>> import pysignalscope as pss
    >>> import numpy as np
    >>> example_waveform = np.array([[0, 1.34, 3.14, 4.48, 6.28],[-175.69, 103.47, 175.69, -103.47,-175.69]])
    >>> out = pss.fft(example_waveform, plot=True, mode='rad', f0=25000, title='ffT input current')

    :param period_vector_t_i: numpy-array [[time-vector[,[current-vector]]. One period only
    :type period_vector_t_i: np.array or List[List[float]
    :param sample_factor: f_sampling/f_period with default = 1000 (optional parameter)
    :type sample_factor: int
    :param plot: insert anything else than "no" or 'False' to show a plot to visualize input and output (optional parameter)
    :type plot: str
    :param mode: 'rad'[default]: full period is 2*pi, 'deg': full period is 360°, 'time': time domain. (optional parameter)
    :type mode: str
    :param f0: fundamental frequency. Needs to be set in 'rad'- or 'deg'-modewith (optional parameter)
    :type f0: float
    :param title: plot window title, defaults to 'ffT' (optional parameter)
    :type title: str
    :param filter_type: 'factor'[default] or 'harmonic' or 'disabled'. (optional parameter)
    :type filter_type: str
    :param filter_value_factor: filters out amplitude-values below a certain factor of max. input amplitude. (optional parameter)
        Should be 0...1, default to 0.01 (1%)
    :type filter_value_factor: float
    :param filter_value_harmonic: filters out harmonics up to a certain number with default = 100 (optional parameter)
        Note: count 1 is DC component, count 2 is the fundamental frequency
    :type filter_value_harmonic: int
    :param figure_size: None for auto-fit; fig_size for matplotlib (width, length) (optional parameter)
    :type figure_size: Tuple
    :param figure_directory: full path with file extension  (optional parameter)
    :type figure_directory: Tuple

    :return: numpy-array [[frequency-vector],[amplitude-vector],[phase-vector]]
    :rtype: npt.NDArray[list]
    """
    # check for correct input parameter
    if (mode == 'rad' or mode == 'deg') and f0 is None:
        raise ValueError("if mode is 'rad' or 'deg', a fundamental frequency f0 must be set")
    # check for input is list. Convert to numpy-array
    if isinstance(period_vector_t_i, list):
        if plot is not False:
            print("Input is list, convert to np.array()")
        period_vector_t_i = np.array(period_vector_t_i)

    # first value of time vector must be zero
    if period_vector_t_i[0][0] != 0:
        raise ValueError("Period vector must start with 0 seconds!")

    # mode pre-calculation
    if mode == 'rad' and f0 is not None:
        period_vector_t_i[0] = period_vector_t_i[0] / (2 * np.pi * f0)
    elif mode == 'deg' and f0 is not None:
        period_vector_t_i[0] = period_vector_t_i[0] / (360 * f0)
    elif mode != 'time':
        raise ValueError("Mode not available. Choose: 'rad', 'deg', 'time'.")

    t = period_vector_t_i[0]
    i = period_vector_t_i[1]

    # fft-function works per default in time domain
    t_interp = np.linspace(0, t[-1], sample_factor)
    i_interp = np.interp(t_interp, t, i)
    if f0 is None:
        f0 = round(1 / t[-1])
        Fs = f0 * sample_factor

    # frequency domain
    f = np.linspace(0, (sample_factor - 1) * f0, sample_factor)
    x = np.fft.fft(i_interp)
    x_mag = np.abs(x) / sample_factor
    print(f"{x_mag=}")
    phi_rad = np.angle(x)

    f_corrected = list(f[0:int(sample_factor / 2 + 1)])
    x_mag_corrected = list(2 * x_mag[0:int(sample_factor / 2 + 1)])
    x_mag_corrected[0] = x_mag_corrected[0] / 2
    phi_rad_corrected = list(phi_rad[0:int(sample_factor / 2 + 1)])

    f_out: List[float] = []
    x_out = []
    phi_rad_out = []
    if filter_type.lower() == 'factor':
        for count, _ in enumerate(x_mag_corrected):
            if x_mag_corrected[count] > filter_value_factor * max(abs(i)):
                f_out.append(f_corrected[count])
                x_out.append(x_mag_corrected[count])
                phi_rad_out.append(phi_rad_corrected[count])
    elif filter_type.lower() == 'harmonic':
        for count, _ in enumerate(x_mag_corrected):
            if count < filter_value_harmonic:
                f_out.append(f_corrected[count])
                x_out.append(x_mag_corrected[count])
                phi_rad_out.append(phi_rad_corrected[count])
    elif filter_type.lower() == 'disabled':
        f_out = f_corrected
        x_out = x_mag_corrected
        phi_rad_out = phi_rad_corrected
    else:
        raise ValueError(
            f"filter_type '{filter_value_harmonic}' not available: Must be 'factor','harmonic' or 'disabled ")

    if plot:
        print(f"{title=}")
        print(f"{t[-1]=}")
        print(f"{f0=}")
        print(f"{Fs=}")
        print(f"{sample_factor=}")
        print(f"f_out = {np.around(f_out, decimals=0)}")
        print(f"x_out = {np.around(x_out, decimals=3)}")
        print(f"phi_rad_out = {np.around(phi_rad_out, decimals=3)}")

        reconstructed_signal = 0
        for i_range in range(len(f_out)):
            reconstructed_signal += x_out[i_range] * np.cos(
                2 * np.pi * f_out[i_range] * t_interp + phi_rad_out[i_range])

        fig, [ax1, ax2, ax3] = plt.subplots(num=title, nrows=3, ncols=1, figsize=figure_size)
        ax1.plot(t, i, label='original signal')
        ax1.plot(t_interp, reconstructed_signal, label='reconstructed signal')
        ax1.grid()
        ax1.set_title('Signal')
        ax1.set_xlabel('time in s')
        ax1.set_ylabel('Amplitude')
        ax1.legend()
        ax2.stem(f_out, x_out)
        ax2.grid()
        ax2.set_title('ffT')
        ax2.set_xlabel('Frequency in Hz')
        ax2.set_ylabel('Amplitude')

        ax3.stem(f_out, phi_rad_out)
        ax3.grid()
        ax3.set_title('ffT')
        ax3.set_xlabel('Frequency in Hz')
        ax3.set_ylabel('Phase in rad')

        plt.tight_layout()
        if figure_directory is not None:
            plt.savefig(figure_directory, bbox_inches="tight")
        plt.show()

    return [f_out, x_out, phi_rad_out]
